__all__ = ('ThirdPartySKU',)

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_distributor, parse_id, parse_sku, put_distributor_into, put_id_into, put_sku_into, validate_distributor,
    validate_id, validate_sku
)


class ThirdPartySKU(RichAttributeErrorBaseType):
    """
    Represents a third party Stock Keeping Unit.
    
    Attributes
    ----------
    distributor : `str`
        The distributor of the SKU.
    id : `str`
        The identifier of the third party SKU.
    sku : `str`
        Stock keeping unit. Might be same as ``.id``.
    """
    __slots__ = ('distributor', 'id', 'sku')
    
    def __new__(cls, *, distributor = ..., sku_id = ..., sku = ...):
        """
        Creates an sku.
        
        Parameters
        ----------
        distributor : `str`, Optional (Keyword only)
            The distributor of the SKU.
        
        sku_id : `str`, Optional (Keyword only)
            The identifier of the third party SKU.
        
        sku : `str`, Optional (Keyword only)
            Stock keeping unit.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # distributor
        if distributor is ...:
            distributor = ''
        else:
            distributor = validate_distributor(distributor)
        
        # id
        if sku_id is ...:
            sku_id = ''
        else:
            sku_id = validate_id(sku_id)
        
        # sku
        if sku is ...:
            sku = ''
        else:
            sku = validate_sku(sku)
        
        self = object.__new__(cls)
        self.distributor = distributor
        self.id = sku_id
        self.sku = sku
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new sku with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Executable data.
        """
        self = object.__new__(cls)
        self.distributor = parse_distributor(data)
        self.id = parse_id(data)
        self.sku = parse_sku(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the sku to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        put_distributor_into(self.distributor, data, defaults)
        put_id_into(self.id, data, defaults)
        put_sku_into(self.sku, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the sku's representation."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        repr_parts.append(', distributor = ')
        repr_parts.append(repr(self.distributor))
        
        repr_parts.append(', id = ')
        repr_parts.append(repr(self.id))
        
        repr_parts.append(', sku = ')
        repr_parts.append(repr(self.sku))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two sku-s are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # distributor
        if self.distributor != other.distributor:
            return False
        
        # id
        if self.id != other.id:
            return False
        
        # sku
        if self.sku != other.sku:
            return False
        
        return True
    
    
    def __gt__(self, other):
        """Returns whether self is greater than the other sku."""
        if type(self) is not type(other):
            return NotImplemented
        
        # distributor
        self_distributor = self.distributor
        other_distributor = other.distributor
        if self_distributor > other_distributor:
            return True
        
        if self_distributor < other_distributor:
            return False
        
        # sku
        self_sku = self.sku
        other_sku = other.sku
        if self_sku > other_sku:
            return True
        
        if self_sku < other_sku:
            return False
        
        # id
        return (self.id > other.id)
    
    
    def __hash__(self):
        """Returns the sku's hash."""
        hash_value = 0
        
        # distributor
        hash_value ^= hash(self.distributor)
        
        # id
        sku_id = self.id
        hash_value ^= hash(sku_id)
        
        # sku
        sku = self.sku
        if sku != sku_id:
            hash_value ^= hash(sku)
        
        return hash_value

    
    def copy(self):
        """
        Copies the sku.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.distributor = self.distributor
        new.id = self.id
        new.sku = self.sku
        return new
    
    
    def copy_with(self, *, distributor = ..., sku_id = ..., sku = ...):
        """
        Copies the sku with the given fields.
        
        Parameters
        ----------
        distributor : `str`, Optional (Keyword only)
            The distributor of the SKU.
        
        sku_id : `str`, Optional (Keyword only)
            The identifier of the third party SKU.
        
        sku : `str`, Optional (Keyword only)
            Stock keeping unit.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # distributor
        if distributor is ...:
            distributor = self.distributor
        else:
            distributor = validate_distributor(distributor)
        
        # id
        if sku_id is ...:
            sku_id = self.id
        else:
            sku_id = validate_id(sku_id)
        
        # sku
        if sku is ...:
            sku = self.sku
        else:
            sku = validate_sku(sku)
        
        new = object.__new__(type(self))
        new.distributor = distributor
        new.id = sku_id
        new.sku = sku
        return new



__all__ = ()

from .....env import API_VERSION

from ....channel import PermissionOverwriteTargetType

from .shared import _convert_preinstanced, convert_deprecated, convert_permission, convert_snowflake


def convert_permission_overwrite_target_type(name, data):
    return _convert_preinstanced('target_type', data, PermissionOverwriteTargetType)


def convert_snowflake__target_id(name, data):
    return convert_snowflake('target_id', data)


CHANNEL_PERMISSION_OVERWRITE_CONVERTERS = {
    'allow': convert_deprecated if API_VERSION in (6, 7) else convert_permission,
    'allow_new': convert_permission if API_VERSION in (6, 7) else convert_deprecated,
    'deny': convert_deprecated if API_VERSION in (6, 7) else convert_permission,
    'deny_new': convert_permission if API_VERSION in (6, 7) else convert_deprecated,
    'id': convert_snowflake__target_id,
    'type': convert_permission_overwrite_target_type,
}



__all__ = ('ChannelMetadataGuildThreadBase', )

from scarletio import copy_docs

from ...core import CHANNELS, GUILDS
from ...permission.permission import PERMISSION_NONE
from ...utils import id_to_datetime

from .constants import AUTO_ARCHIVE_DEFAULT, SLOWMODE_DEFAULT
from .fields import (
    parse_archived, parse_archived_at, parse_auto_archive_after, parse_created_at, parse_open, parse_owner_id,
    parse_slowmode, put_archived_at_into, put_archived_into, put_auto_archive_after_into, put_created_at_into,
    put_open_into, put_owner_id_into, put_slowmode_into, validate_archived, validate_archived_at,
    validate_auto_archive_after, validate_created_at, validate_open, validate_owner_id, validate_slowmode
)

from .guild_base import ChannelMetadataGuildBase


class ChannelMetadataGuildThreadBase(ChannelMetadataGuildBase):
    """
    Base guild channel metadata type.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    _created_at : `None`, `datetime`
        When the channel was created.
    archived : `bool`
        Whether the thread is archived.
    archived_at : `None`, `datetime`
        When the thread's archive status was last changed.
    auto_archive_after : `int`
        Duration in seconds to automatically archive the thread after recent activity. Can be one of: `3600`, `86400`,
        `259200`, `604800`.
    open : `bool`
        Whether the thread channel is open.
    slowmode : `int`
        The amount of time in seconds what a user needs to wait between it's each message. Bots and user accounts with
        `manage_messages`, `manage_channels` permissions are unaffected.
    thread_users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        The users inside of the thread if any.
    owner_id : `int`
        The channel's creator's identifier. Defaults to `0`.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = (
        '_created_at', 'archived', 'archived_at', 'auto_archive_after', 'open', 'owner_id', 'slowmode', 'thread_users'
    )
    
    
    @copy_docs(ChannelMetadataGuildBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataGuildBase.__hash__(self)
        
        # created_at
        created_at = self._created_at
        if (created_at is not None):
            hash_value ^= hash(created_at)
        
        # archived
        hash_value ^= self.archived << 5
        
        # archived_at
        hash_value ^= hash(self.archived_at)
        
        # open
        hash_value ^= self.open << 7
        
        # owner_id
        hash_value ^= self.owner_id
        
        # slowmode
        hash_value ^= self.slowmode << 4
        
        return hash_value
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildBase.from_data)
    def from_data(cls, data):
        self = super(ChannelMetadataGuildThreadBase, cls).from_data(data)

        # created_at
        self._created_at = parse_created_at(data)
        
        # owner_id
        self.owner_id = parse_owner_id(data)
        
        # thread_users
        self.thread_users = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildBase._created)
    def _created(self, channel_entity, client):
        try:
            guild = GUILDS[channel_entity.guild_id]
        except KeyError:
            pass
        else:
            guild.threads[channel_entity.id] = channel_entity
    
    
    @copy_docs(ChannelMetadataGuildBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not ChannelMetadataGuildBase._is_equal_same_type(self, other):
            return False
        
        # archived
        if self.archived != other.archived:
            return False
        
        # archived_at
        if self.archived_at != other.archived_at:
            return False
        
        # auto_archive_after
        if self.auto_archive_after != other.auto_archive_after:
            return False
        
        # created_at
        if self._created_at != other._created_at:
            return False
        
        # open
        if self.open != other.open:
            return False
        
        # owner_id
        if self.owner_id != other.owner_id:
            return False
        
        # slowmode
        if self.slowmode != other.slowmode:
            return False
        
        # thread_users
        # Ignore this field
        
        return True
    
    
    @classmethod
    @copy_docs(ChannelMetadataGuildBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildThreadBase, cls)._create_empty()
        
        self._created_at = None
        self.archived = False
        self.archived_at = None
        self.auto_archive_after = AUTO_ARCHIVE_DEFAULT
        self.open = False
        self.owner_id = 0
        self.slowmode = SLOWMODE_DEFAULT
        self.thread_users = None
        
        return self
    
    
    @copy_docs(ChannelMetadataGuildBase._get_display_name)
    def _get_display_name(self):
        return self.name.lower()
    
    
    @copy_docs(ChannelMetadataGuildBase._get_created_at)
    def _get_created_at(self, channel_entity):
        created_at = self._created_at
        if (created_at is None):
            created_at = id_to_datetime(channel_entity.id)
        
        return created_at
    
    
    @copy_docs(ChannelMetadataGuildBase._get_users)
    def _get_users(self, channel_entity):
        thread_users = self.thread_users
        if thread_users is None:
            users = []
        else:
            users = list(thread_users.values())
        
        return users
    
    
    @copy_docs(ChannelMetadataGuildBase._iter_users)
    def _iter_users(self, channel_entity):
        thread_users = self.thread_users
        if (thread_users is not None):
            yield from thread_users.values()
    
    
    @copy_docs(ChannelMetadataGuildBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataGuildBase._update_attributes(self, data)
        
        # archived
        self.archived = parse_archived(data)
        
        # archived_at
        self.archived_at = parse_archived_at(data)
        
        # auto_archive_after
        self.auto_archive_after = parse_auto_archive_after(data)
        
        # open
        self.open = parse_open(data)
        
        # slowmode
        self.slowmode = parse_slowmode(data)
    
    
    @copy_docs(ChannelMetadataGuildBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataGuildBase._difference_update_attributes(self, data)
        
        # archived
        archived = parse_archived(data)
        if (self.archived != archived):
            old_attributes['archived'] = self.archived
            self.archived = archived
        
        # archived_at
        archived_at = parse_archived_at(data)
        if (self.archived_at != archived_at):
            old_attributes['archived_at'] = self.archived_at
            self.archived_at = archived_at
        
        # auto_archive_after
        auto_archive_after = parse_auto_archive_after(data)
        if (self.auto_archive_after != auto_archive_after):
            old_attributes['auto_archive_after'] = self.auto_archive_after
            self.auto_archive_after = auto_archive_after
        
        # open
        open_ = parse_open(data)
        if (self.open != open_):
            old_attributes['open'] = self.open
            self.open = open_
        
        # slowmode
        slowmode = parse_slowmode(data)
        if self.slowmode != slowmode:
            old_attributes['slowmode'] = self.slowmode
            self.slowmode = slowmode
        
        return old_attributes
    
    
    @copy_docs(ChannelMetadataGuildBase._delete)
    def _delete(self, channel_entity, client):
        guild_id = channel_entity.guild_id
        try:
            guild = GUILDS[guild_id]
        except KeyError:
            pass
        else:
            try:
                del guild.threads[channel_entity.id]
            except KeyError:
                pass
        
        thread_users = self.thread_users
        if (thread_users is not None):
            self.thread_users = None
            channel_id = channel_entity.id
            for user in thread_users.values():
                thread_profiles = user.thread_profiles
                if (thread_profiles is not None):
                    try:
                        del thread_profiles[channel_id]
                    except KeyError:
                        pass
                    else:
                        if (not thread_profiles):
                            user.thread_profiles = None
    
    
    @copy_docs(ChannelMetadataGuildBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        try:
            parent = CHANNELS[self.parent_id]
        except KeyError:
            return PERMISSION_NONE
        
        return parent.permissions_for(user)
    
    
    @copy_docs(ChannelMetadataGuildBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        try:
            parent = CHANNELS[self.parent_id]
        except KeyError:
            return PERMISSION_NONE
        
        return parent.permissions_for_roles(*roles)
    
    
    @copy_docs(ChannelMetadataGuildBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        ChannelMetadataGuildBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
        
        # archived
        try:
            archived = keyword_parameters.pop('archived')
        except KeyError:
            pass
        else:
            self.archived = validate_archived(archived)
        
        # archived_at
        try:
            archived_at = keyword_parameters.pop('archived_at')
        except KeyError:
            pass
        else:
            self.archived_at = validate_archived_at(archived_at)
        
        # auto_archive_after
        try:
            auto_archive_after = keyword_parameters.pop('auto_archive_after')
        except KeyError:
            pass
        else:
            self.auto_archive_after = validate_auto_archive_after(auto_archive_after)
        
        # created_at
        try:
            created_at = keyword_parameters.pop('created_at')
        except KeyError:
            pass
        else:
            self._created_at = validate_created_at(created_at)
        
        # open
        try:
            open_ = keyword_parameters.pop('open')
        except KeyError:
            pass
        else:
            self.open = validate_open(open_)
        
        # owner_id
        try:
            owner_id = keyword_parameters.pop('owner_id')
        except KeyError:
            pass
        else:
            self.owner_id = validate_owner_id(owner_id)
        
        # slowmode
        try:
            slowmode = keyword_parameters.pop('slowmode')
        except KeyError:
            pass
        else:
            self.slowmode = validate_slowmode(slowmode)
    
    
    @copy_docs(ChannelMetadataGuildBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataGuildBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # archived
        if include_internals:
            put_archived_into(self.archived, data, defaults)
        
        # archived_at
        if include_internals:
            put_archived_at_into(self.archived_at, data, defaults)
        
        # auto_archive_after
        put_auto_archive_after_into(
            self.auto_archive_after, data, defaults, flatten_thread_metadata = not include_internals
        )
        
        # created_at
        if include_internals:
            put_created_at_into(self._created_at, data, defaults)
        
        # open
        put_open_into(self.open, data, defaults, flatten_thread_metadata = not include_internals)
        
        # owner_id
        if include_internals:
            put_owner_id_into(self.owner_id, data, defaults)
        
        # slowmode
        put_slowmode_into(self.slowmode, data, defaults)
        
        return data



__all__ = ('ActivityTimestamps',)

from datetime import datetime

from scarletio import copy_docs

from ...utils import DATETIME_FORMAT_CODE, datetime_to_millisecond_unix_time, millisecond_unix_time_to_datetime

from .base import ActivityFieldBase


def _assert__activity_timestamp__start(start):
    """
    Asserts the `start` parameter of ``ActivityTimestamps.__new__`` method.
    
    Parameters
    ----------
    start : `None`, `datetime`
       When the activity starts.
    
    Raises
    ------
    AssertionError
        - If `start` is not `None`, `datetime`.
    """
    if (start is not None) and (not isinstance(start, datetime)):
        raise AssertionError(
            f'`start` can be `None`, `datetime`, got {start.__class__.__name__}; {start!r}.'
        )
    
    return True


def _assert__activity_timestamp__end(end):
    """
    Asserts the `end` parameter of ``ActivityTimestamps.__new__`` method.
    
    Parameters
    ----------
    end : `None`, `datetime`
        When the activity ends.
    
    Raises
    ------
    AssertionError
        - If `end` is not `None`, `datetime`.
    """
    if (end is not None) and (not isinstance(end, datetime)):
        raise AssertionError(
            f'`end` can be `None`, `datetime`, got {end.__class__.__name__}; {end!r}.'
        )
    
    return True


class ActivityTimestamps(ActivityFieldBase):
    """
    Represents an activity's timestamp field.
    
    Attributes
    ----------
    end : `None`, `datetime`
        When the activity ends.
    start : `None`, `datetime`
       When the activity starts.
    """
    __slots__ = ('end', 'start',)
    
    def __new__(cls, *, start=None, end=None):
        """
        Creates a new activity timestamp with the given parameters.
        
        Parameters
        ----------
        end : `None`, `datetime` = `None`, Optional (Keyword only)
            When the activity ends. 
        start : `None`, `datetime` = `None`, Optional (Keyword only)
           When the activity starts.
        """
        assert _assert__activity_timestamp__start(start)
        assert _assert__activity_timestamp__end(end)
        
        self = object.__new__(cls)
        self.start = start
        self.end = end
        return self
    
    
    @classmethod
    @copy_docs(ActivityFieldBase.from_data)
    def from_data(cls, timestamps_data):
        start = timestamps_data.get('start', None)
        if (start is not None):
            start = millisecond_unix_time_to_datetime(start)
        
        end = timestamps_data.get('end', None)
        if (end is not None):
            end = millisecond_unix_time_to_datetime(end)
        
        self = object.__new__(cls)
        self.start = start
        self.end = end
        return self
    
    
    @copy_docs(ActivityFieldBase.to_data)
    def to_data(self):
        timestamps_data = {}
        
        start = self.start
        if (start is not None):
            timestamps_data['start'] = datetime_to_millisecond_unix_time(start)
        
        end = self.end
        if (end is not None):
            timestamps_data['end'] = datetime_to_millisecond_unix_time(end)
        
        return timestamps_data
    
    
    @copy_docs(ActivityFieldBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        start = self.start
        if (start is not None):
            repr_parts.append(' start=')
            repr_parts.append(start.__format__(DATETIME_FORMAT_CODE))
            field_added = True
        else:
            field_added = False
        
        end = self.end
        if (end is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' end=')
            repr_parts.append(start.__format__(DATETIME_FORMAT_CODE))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityFieldBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.start != other.start:
            return False
        
        if self.end != other.end:
            return False
        
        return True
    
    
    @copy_docs(ActivityFieldBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        start = self.start
        if (start is not None):
            hash_value ^= hash(start)
            hash_value ^= (1 << 0)
        
        end = self.end
        if (end is not None):
            hash_value ^= hash(end)
            hash_value ^= (1 << 4)
        
        return hash_value


    @copy_docs(ActivityFieldBase.__bool__)
    def __bool__(self):
        start = self.start
        if (start is not None):
            return True
        
        end = self.end
        if (end is not None):
            return True
        
        return False



from .base import *
from .constants import *
from .fields import *
from .keyword import *
from .keyword_preset import *
from .mention_spam import *
from .preinstanced import *
from .utils import *


__all__ = (
    *base.__all__,
    *constants.__all__,
    *fields.__all__,
    *keyword.__all__,
    *keyword_preset.__all__,
    *mention_spam.__all__,
    *preinstanced.__all__,
    *utils.__all__,
)



from .helpers import *
from .slash_command import *
from .slash_command_category import *
from .slash_command_function import *
from .slash_command_parameter_auto_completer import *

__all__ = (
    *helpers.__all__,
    *slash_command.__all__,
    *slash_command_category.__all__,
    *slash_command_function.__all__,
    *slash_command_parameter_auto_completer.__all__,
)



__all__ = ()

from ..field_parsers import preinstanced_array_parser_factory, int_parser_factory
from ..field_putters import preinstanced_array_putter_factory, int_putter_factory
from ..field_validators import preinstanced_array_validator_factory, int_options_validator_factory

from .preinstanced import GuildFeature

# features

parse_features = preinstanced_array_parser_factory('features', GuildFeature)
put_features_into = preinstanced_array_putter_factory('features')
validate_features = preinstanced_array_validator_factory('features', GuildFeature)

# premium_tier

parse_premium_tier = int_parser_factory('premium_tier', 0)
put_premium_tier_into = int_putter_factory('premium_tier')
validate_premium_tier = int_options_validator_factory('premium_tier', frozenset((range(4))))



from .fields import *
from .metadata import *

from .activity import *
from .constants import *
from .flags import *
from .preinstanced import *

__all__ = (
    *fields.__all__,
    *metadata.__all__,
        
    *activity.__all__,
    *constants.__all__,
    *flags.__all__,
    *preinstanced.__all__,
)

# Construct deprecations

from scarletio import modulize

from ...utils.module_deprecation import deprecated_import


@deprecated_import
@modulize
class ACTIVITY_TYPES:
    """
    Deprecated. Please use ``ActivityType`` instead.
    """
    game = 0
    stream = 1
    spotify = 2
    watching = 3
    custom = 4
    competing = 5


deprecated_import(Activity, 'ActivityRich')



__all__ = ('SlashCommand',)

from functools import partial as partial_func

from scarletio import WeakReferer, copy_docs, export

from .....discord.application_command import ApplicationCommandTargetType
from .....discord.application_command.constants import APPLICATION_COMMAND_OPTIONS_MAX
from .....discord.events.handling_helpers import Router, _EventHandlerManager, check_name, route_name, route_value

from ...converters import get_slash_command_parameter_converters
from ...exceptions import SlashCommandParameterConversionError, handle_command_exception
from ...response_modifier import ResponseModifier
from ...utils import _check_maybe_route, raw_name_to_display
from ...wrappers import CommandWrapper, get_parameter_configurers

from ..command_base_application_command import CommandBaseApplicationCommand
from ..command_base_application_command.constants import (
    APPLICATION_COMMAND_CATEGORY_DEEPNESS_START, APPLICATION_COMMAND_DEEPNESS,
    APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND, APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY
)
from ..command_base_application_command.helpers import (
    _reset_application_command_schema, _validate_allow_by_default, _validate_allow_in_dm, _validate_delete_on_unload,
    _validate_guild, _validate_is_global, _validate_name, _validate_nsfw, _validate_required_permissions
)


from .helpers import (
    _build_auto_complete_parameter_names, _generate_description_from, _register_auto_complete_function,
    _validate_is_default
)
from .slash_command_category import SlashCommandCategory
from .slash_command_function import SlashCommandFunction
from .slash_command_parameter_auto_completer import SlashCommandParameterAutoCompleter


@export
class SlashCommand(CommandBaseApplicationCommand):
    """
    Class to wrap an application command providing interface for ``Slasher``.
    
    Attributes
    ----------
    _exception_handlers : `None`, `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
        
        Same as ``Slasher._exception_handlers``.
    
    _parent_reference : `None`, ``WeakReferer`` to ``Slasher``
        Reference to the slasher application command's parent.
    
    name : `str`
        Application command name. It's length can be in range [1:32].
    
    _permission_overwrites : `None`, `dict` of (`int`, `list` of ``ApplicationCommandPermissionOverwrite``)
        Permission overwrites applied to the slash command.
    
    _registered_application_command_ids : `None`, `dict` of (`int`, `int`) items
        The registered application command ids, which are matched by the command's schema.
        
        If empty set as `None`, if not then the keys are the respective guild's id and the values are the application
        command id.
    
    _schema : `None`, ``ApplicationCommand``
        Internal slot used by the ``.get_schema`` method.
    
    _unloading_behaviour : `int`
        Behaviour what describes what should happen when the command is removed from the slasher.
        
        Can be any of the following:
        
        +-------------------------------+-------+
        | Respective name               | Value |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_DELETE    | 0     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_KEEP      | 1     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_INHERIT   | 2     |
        +-------------------------------+-------+
    
    allow_by_default : `None`, `bool`
        Whether the command is enabled by default for everyone who has `use_application_commands` permission.
    
    allow_in_dm : `None`, `bool`
        Whether the command can be used in private channels (dm).
    
    global_ : `bool`
        Whether the command is a global command.
        
        Global commands have their ``.guild_ids`` set as `None`.
    
    guild_ids : `None`, `set` of `int`
        The ``Guild``'s id to which the command is bound to.
    
    nsfw : `None,` bool`
        Whether the application command is only allowed in nsfw channels.
    
    required_permissions : `None`, ``Permission``
        The required permissions to use the application command inside of a guild.
    
    _auto_completers : `None`, `list` of ``SlashCommandParameterAutoCompleter``
        Auto completer functions.
    
    _command : `None`, ``SlashCommandFunction``
        The command of the slash command.
    
    _self_reference : `None`, ``WeakReferer`` to ``SlashCommand``
        Back reference to the slasher application command.
        
        Used by sub commands to access the parent entity.
    
    _sub_commands : `None`, `dict` of (`str`, (``SlashCommandFunction`` or ``SlashCommandCategory``)) items
        Sub-commands of the slash command.
        
        Mutually exclusive with the ``._command`` parameter.
    
    default : `bool`
        Whether the command is the default command in it's category.
    
    description : `str`
        Application command description. It's length can be in range [2:100].
    
    
    Class Attributes
    ----------------
    COMMAND_COMMAND_NAME : `str`
        The command's name defining parameter's name.
    COMMAND_PARAMETER_NAMES : `tuple` of `str`
        All parameters names accepted by ``.__new__``
    COMMAND_NAME_NAME : `str`
        The command's "command" defining parameter's name.
    
    Notes
    -----
    ``SlashCommand``-s are weakreferable.
    """
    __slots__ = (
        '__weakref__', '_auto_completers', '_command', '_self_reference', '_sub_commands', 'description'
    )
    
    SLASH_COMMAND_PARAMETER_NAMES = (
        *CommandBaseApplicationCommand.COMMAND_PARAMETER_NAMES,
        'description',
        'is_default',
    )
    
    @CommandBaseApplicationCommand.target.getter
    def target(self):
        return ApplicationCommandTargetType.chat
    
    
    def __new__(
        cls, func, name=None, description = None, is_global = None, guild=None, is_default = None, delete_on_unload=None,
        allow_by_default = None, allow_in_dm=None, required_permissions = None, nsfw = None, **keyword_parameters,
    ):
        """
        Creates a new ``SlashCommand`` with the given parameters.
        
        Parameters
        ----------
        func : `None`, `async-callable` = `None`, Optional
            The function used as the command when using the respective slash command.
        
        name : `None`, `str`, `tuple` of (`str`, `Ellipsis`, `None`) = `None`, Optional
            The command's name if applicable. If not given or if given as `None`, the `func`'s name will be use
            instead.
        
        description : `None`, `Any`, `tuple` of (`None`, `Ellipsis`, `Any`) = `None`, Optional
            Description to use instead of the function's docstring.
        
        is_global : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the slash command is global. Defaults to `False`.
        
        guild : `None`, ``Guild``,  `int`, (`list`, `set`) of (`int`, ``Guild``) or \
                `tuple` of (`None`, ``Guild``,  `int`, `Ellipsis`, (`list`, `set`) of (`int`, ``Guild``)) = `None`
                , Optional
            To which guild(s) the command is bound to.
        
        is_global : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the slash command is the default command in it's category.
        
        delete_on_unload : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the command should be deleted from Discord when removed.
        
        allow_by_default : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the command is enabled by default for everyone who has `use_application_commands` permission.
        
        allow_in_dm : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the command can be used in private channels (dm).
        
        required_permissions : `None`, `int`, ``Permission``, `tuple` of (`None`, `int`, ``Permission``,
                `Ellipsis`) = `None`, Optional
            The required permissions to use the application command inside of a guild.
        
        nsfw : `None`, `bool`, `tuple` of (`None`, `bool`, `Ellipsis`) = `None`, Optional
            Whether the application command is only allowed in nsfw channels.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other parameters
        ----------------
        allowed_mentions : `None`, `str`, ``UserBase``, ``Role``, ``AllowedMentionProxy``, \
                `list` of (`str`, ``UserBase``, ``Role`` ), Optional (Keyword only)
            Which user or role can the response message ping (or everyone).
        show_for_invoking_user_only : `bool`, Optional (Keyword only)
            Whether the response message should only be shown for the invoking user.
        wait_for_acknowledgement : `bool`, Optional (Keyword only)
            Whether acknowledge tasks should be ensure asynchronously.
        
        Returns
        -------
        self : ``SlashCommand``, ``Router``
        
        Raises
        ------
        TypeError
            If a parameter's type is incorrect.
        ValueError
            If a parameter's value is incorrect.
        """
        if (func is not None) and isinstance(func, CommandWrapper):
            command, wrappers = func.fetch_function_and_wrappers_back()
        else:
            command = func
            wrappers = None
        
        # Check for routing
        
        route_to = 0
        name, route_to = _check_maybe_route('name', name, route_to, _validate_name)
        description, route_to = _check_maybe_route('description', description, route_to, None)
        is_global, route_to = _check_maybe_route('is_global', is_global, route_to, _validate_is_global)
        guild_ids, route_to = _check_maybe_route('guild', guild, route_to, _validate_guild)
        is_default, route_to = _check_maybe_route('is_default', is_default, route_to, _validate_is_default)
        unloading_behaviour, route_to = _check_maybe_route(
            'delete_on_unload', delete_on_unload, route_to, _validate_delete_on_unload
        )
        allow_by_default, route_to = _check_maybe_route(
            'allow_by_default', allow_by_default, route_to, _validate_allow_by_default
        )
        allow_in_dm, route_to = _check_maybe_route('allow_in_dm', allow_in_dm, route_to, _validate_allow_in_dm)
        nsfw, route_to = _check_maybe_route('nsfw', nsfw, route_to, _validate_nsfw)
        required_permissions, route_to = _check_maybe_route(
            'required_permissions', required_permissions, route_to, _validate_required_permissions
        )
        
        if route_to:
            name = route_name(name, route_to)
            name = [raw_name_to_display(check_name(command, sub_name)) for sub_name in name]
            
            default_description = _generate_description_from(command, None, None)
            is_global = route_value(is_global, route_to)
            guild_ids = route_value(guild_ids, route_to)
            is_default = route_value(is_default, route_to)
            unloading_behaviour = route_value(unloading_behaviour, route_to)
            allow_by_default = route_value(allow_by_default, route_to)
            allow_in_dm = route_value(allow_in_dm, route_to)
            nsfw = route_value(nsfw, route_to)
            required_permissions = route_value(required_permissions, route_to)
            
            description = [
                (
                    _generate_description_from(command, sub_name, description)
                    if ((description is None) or (description is not default_description)) else
                    default_description
                )
                for sub_name, description in zip(name, description)
            ]
        
        else:
            name = check_name(command, name)
            name = raw_name_to_display(name)
            
            description = _generate_description_from(command, name, description)
        
        
        # Check extra parameters
        response_modifier = ResponseModifier(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused parameters: {keyword_parameters!r}.'
            )
        
        
        if command is None:
            parameter_converters = None
        else:
            parameter_configurers = get_parameter_configurers(wrappers)
            command, parameter_converters = get_slash_command_parameter_converters(command, parameter_configurers)
    
        
        if route_to:
            router = []
            
            for (
                name, description, is_global, guild_ids, is_default, unloading_behaviour, allow_by_default,
                nsfw, required_permissions, allow_in_dm
            ) in zip(
                name, description, is_global, guild_ids, is_default, unloading_behaviour, allow_by_default,
                nsfw, required_permissions, allow_in_dm
            ):
                
                if is_global and (guild_ids is not None):
                    raise TypeError(
                        f'`is_global` and `guild` contradict each other, got is_global = {is_global!r}, '
                        f'guild={guild!r}.'
                    )
                
                if (command is None):
                    command_function = None
                    sub_commands = {}
                else:
                    command_function = SlashCommandFunction(
                        command, parameter_converters, name, description, response_modifier, is_default
                    )
                    sub_commands = None
                
                self = object.__new__(cls)
                self._command = command_function
                self._sub_commands = sub_commands
                self.description = description
                self.guild_ids = guild_ids
                self.global_ = is_global
                self.name = name
                self._schema = None
                self._registered_application_command_ids = None
                self.default = is_default
                self._unloading_behaviour = unloading_behaviour
                self.allow_by_default = allow_by_default
                self.allow_in_dm = allow_in_dm
                self.nsfw = nsfw
                self.required_permissions = required_permissions
                self._permission_overwrites = None
                self._auto_completers = None
                self._exception_handlers = None
                self._parent_reference = None
                self._self_reference = None
                
                if (command_function is not None):
                    command_function._parent_reference = self._get_self_reference()
                
                if (wrappers is not None):
                    for wrapper in wrappers:
                        wrapper.apply(self)
                
                router.append(self)
            
            return Router(router)
        else:
            if is_global and (guild_ids is not None):
                raise TypeError(
                    f'`is_global` and `guild` contradict each other, got is_global = {is_global!r}, '
                    f'guild={guild!r}.'
                )
            
            if (command is None):
                sub_commands = {}
                command_function = None
            else:
                command_function = SlashCommandFunction(
                    command, parameter_converters, name, description, response_modifier, is_default
                )
                sub_commands = None
            
            self = object.__new__(cls)
            self._command = command_function
            self._sub_commands = sub_commands
            self.description = description
            self.guild_ids = guild_ids
            self.global_ = is_global
            self.name = name
            self._schema = None
            self._registered_application_command_ids = None
            self.default = is_default
            self._unloading_behaviour = unloading_behaviour
            self.allow_by_default = allow_by_default
            self.allow_in_dm = allow_in_dm
            self.nsfw = nsfw
            self.required_permissions = required_permissions
            self._permission_overwrites = None
            self._auto_completers = None
            self._exception_handlers = None
            self._parent_reference = None
            self._self_reference = None
            
            if (command_function is not None):
                command_function._parent_reference = self._get_self_reference()
            
            if (wrappers is not None):
                for wrapper in wrappers:
                    wrapper.apply(self)
            
            return self
    
    
    @copy_docs(CommandBaseApplicationCommand._cursed_repr_builder)
    def _cursed_repr_builder(self):
        for repr_parts in CommandBaseApplicationCommand._cursed_repr_builder(self):
            
            yield repr_parts
            
            description = self.description
            if self.name != description:
                 repr_parts.append(', description = ')
                 repr_parts.append(repr(description))
    
    
    @copy_docs(CommandBaseApplicationCommand.invoke)
    async def invoke(self, client, interaction_event):
        options = interaction_event.interaction.options
        
        command = self._command
        if (command is not None):
            await command.invoke(client, interaction_event, options)
            return
        
        if (options is None) or (len(options) != 1):
            return
        
        option = options[0]
        
        try:
            sub_command = self._sub_commands[option.name]
        except KeyError:
            pass
        else:
            await sub_command.invoke(client, interaction_event, option.options)
            return
        
        # Do not put this into the `except` branch.
        await handle_command_exception(
            self,
            client,
            interaction_event,
            SlashCommandParameterConversionError(
                None,
                option.name,
                'sub-command',
                list(self._sub_commands.keys()),
            )
        )
        return
    
    
    @copy_docs(CommandBaseApplicationCommand.copy)
    def copy(self):
        new = CommandBaseApplicationCommand.copy(self)
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            auto_completers = auto_completers.copy()
        new._auto_completers = auto_completers
        
        # _command
        command = self._command
        if (command is not None):
            command = command.copy()
        new._command = command
        
        # _self_reference
        new._self_reference = None
        
        # _sub_commands
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            sub_commands = {category_name: category.copy() for category_name, category in sub_commands.items()}
        new._sub_commands = sub_commands
        
        # default
        new.default = self.default
        
        # description
        new.description = self.description
        
        # ---- POST LINKING ----
        
        if (sub_commands is not None):
            for sub_command in sub_commands.values():
                sub_command._parent_reference = new._get_self_reference()
        
        return new
    
    
    
    @copy_docs(CommandBaseApplicationCommand.__hash__)
    def __hash__(self):
        hash_value = CommandBaseApplicationCommand.__hash__(self)
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            hash_value ^= len(auto_completers) << 22
            
            for auto_completer in auto_completers:
                hash_value ^= hash(auto_completer)
        
        # _command
        command = self._command
        if (command is not None):
            hash_value ^= hash(command)
        
        # _sub_commands
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            hash_value ^= len(sub_commands) << 26
            
            for sub_command in sub_commands:
                hash_value ^= hash(sub_command)
        
        # default
        hash_value ^= self.default << 30
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        
        return hash_value
    
    
    @copy_docs(CommandBaseApplicationCommand._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not CommandBaseApplicationCommand._is_equal_same_type(self, other):
            return False
        
        # _auto_completers
        if self._auto_completers != other._auto_completers:
            return False
        
        # _command
        if self._command != other._command:
            return False
        
        # _self_reference
        # Internal field
        
        # _sub_commands
        if self._sub_commands != other._sub_commands:
            return False
        
        # default
        if self.default != other.default:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        return True
    
    
    async def invoke_auto_completion(self, client, interaction_event, auto_complete_option):
        """
        Calls the respective auto completion function of the command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        auto_complete_option : ``ApplicationCommandAutocompleteInteraction``, \
                ``ApplicationCommandAutocompleteInteractionOption``
            The option to autocomplete.
        """
        command_function = self._command
        if (command_function is not None):
            await command_function.invoke_auto_completion(client, interaction_event, auto_complete_option)
            return
        
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            auto_complete_option = auto_complete_option.options[0]
            
            auto_complete_option_type = auto_complete_option.type
            if (
                (auto_complete_option_type is APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND) or
                (auto_complete_option_type is APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY)
            ):
                options = auto_complete_option.options
                if (options is not None):
                    try:
                        sub_command = sub_commands[auto_complete_option.name]
                    except KeyError:
                        pass
                    else:
                        await sub_command.invoke_auto_completion(client, interaction_event, auto_complete_option)
            
            return
        
        # no more cases
    
    @copy_docs(CommandBaseApplicationCommand._get_schema_options)
    def _get_schema_options(self):
        command = self._command
        if command is None:
            sub_commands = self._sub_commands
            options = [sub_command.as_option() for sub_command in sub_commands.values()]
        
        else:
            parameter_converters = command._parameter_converters
            
            options = None
            for parameter_converter in parameter_converters:
                option = parameter_converter.as_option()
                if (option is not None):
                    if (options is None):
                        options = []
                    
                    options.append(option)
    
        return options
    
    
    def as_sub_command(self, deepness):
        """
        Returns the slash command as a sub-command or sub-category.
        
        Parameters
        ----------
        deepness : `int`
            How nested the category or function will be.
        
        Returns
        -------
        new : ``SlashCommandFunction``, ``SlashCommandCategory``
        """
        command = self._command
        if (command is not None):
            return command
        
        return SlashCommandCategory(self, deepness)
    
    
    @property
    @copy_docs(CommandBaseApplicationCommand.interactions)
    def interactions(self):
        if self._command is not None:
            raise RuntimeError(
                f'The {self.__class__.__name__} is not a category.'
            )
        
        return _EventHandlerManager(self)
    
    
    def create_event(self, func, *args, **keyword_parameters):
        """
        Adds a sub-command under the slash command.
        
        Parameters
        ----------
        func : `async-callable`
            The function used as the command when using the respective slash command.
        *args : Positional Parameters
            Positional parameters to pass to ``SlashCommand``'s constructor.
        **keyword_parameters : Keyword parameters
            Keyword parameters to pass to the ``SlashCommand``'s constructor.
        
        Returns
        -------
        self : ``SlashCommandFunction``, ``SlashCommandCategory``
        
        Raises
        ------
        TypeError
            If Any parameter's type is incorrect.
        ValueError
            If Any parameter's value is incorrect.
        RuntimeError
            - The ``SlashCommand`` is not a category.
            - The ``SlashCommand`` reached the maximal amount of children.
            - If the command to add is a default sub-command meanwhile the category already has one.
        """
        if self._command is not None:
            raise RuntimeError(f'The {self!r} is not a category.')
        
        if isinstance(func, Router):
            func = func[0]
        
        if isinstance(func, type(self)):
            self._add_application_command(func)
            return self
        
        command = type(self)(func, *args, **keyword_parameters)
        
        if isinstance(command, Router):
            command = command[0]
        
        return self._add_application_command(command)
    
    
    def create_event_from_class(self, klass):
        """
        Breaks down the given class to it's class attributes and tries to add it as a sub-command or sub-category.
        
        Parameters
        ----------
        klass : `type`
            The class, from what's attributes the command will be created.
        
        Returns
        -------
        self : ``SlashCommandFunction``, ``SlashCommandCategory``
         
        Raises
        ------
        TypeError
            If Any attribute's type is incorrect.
        ValueError
            If Any attribute's value is incorrect.
        RuntimeError
            - The ``SlashCommand`` is not a category.
            - The ``SlashCommand`` reached the maximal amount of children.
            - If the command to add is a default sub-command meanwhile the category already has one.
        """
        command = type(self).from_class(klass)
        
        if isinstance(command, Router):
            command = command[0]
        
        return self._add_application_command(command)
    
    
    def _add_application_command(self, command):
        """
        Adds a sub-command or sub-category to the slash command.
        
        Parameters
        ----------
        command : ``SlashCommand``
            The slash command to add.
        
        Raises
        ------
        RuntimeError
            - The ``SlashCommand`` reached the maximal amount of children.
            - If the command to add is a default sub-command meanwhile the category already has one.
        """
        sub_commands = self._sub_commands
        if len(sub_commands) == APPLICATION_COMMAND_OPTIONS_MAX and (command.name not in sub_commands):
            raise RuntimeError(
                f'The {self!r} reached the maximal amount of children '
                f'({APPLICATION_COMMAND_OPTIONS_MAX}).'
            )
        
        if command.default:
            for sub_command in sub_commands.values():
                if sub_command.default:
                    raise RuntimeError(
                        f'{self!r} already has a default command.'
                    )
        
        as_sub_command = command.as_sub_command(APPLICATION_COMMAND_CATEGORY_DEEPNESS_START)
        as_sub_command._parent_reference = self._get_self_reference()
        
        sub_commands[as_sub_command.name] = as_sub_command
        _reset_application_command_schema(self)
        
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            for auto_completer in auto_completers:
                as_sub_command._try_resolve_auto_completer(auto_completer)
        
        return as_sub_command
    
    
    @copy_docs(CommandBaseApplicationCommand.get_real_command_count)
    def get_real_command_count(self):
        if (self._command is None):
            sub_commands = self._sub_commands
            real_command_count = 0
            
            if (sub_commands is not None):
                for sub_command_or_category in sub_commands.values():
                    if isinstance(sub_command_or_category, SlashCommandFunction):
                        real_command_count += 1
                    else:
                        # Nesting more is not allowed by Discord.
                        real_command_count += len(sub_command_or_category._sub_commands)
        
        else:
            real_command_count = 1
        
        return real_command_count
    
    
    @copy_docs(CommandBaseApplicationCommand.autocomplete)
    def autocomplete(self, parameter_name, *parameter_names, function=None):
        parameter_names = _build_auto_complete_parameter_names(parameter_name, parameter_names)
        
        if (function is None):
            return partial_func(_register_auto_complete_function, self, parameter_names)
        
        return self._add_autocomplete_function(parameter_names, function)
    
    
    def _add_autocomplete_function(self, parameter_names, function):
        """
        Registers an autocomplete function.
        
        Parameters
        ----------
        parameter_names : `list` of `str`
            The parameters' names.
        function : `async-callable`
            The function to register as auto completer.
        
        Returns
        -------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The registered auto completer
        
        Raises
        ------
        RuntimeError
            - If the application command function has no parameter named, like `parameter_name`.
            - If the parameter cannot be auto completed.
        TypeError
            - If `function` is not an asynchronous.
        """
        if isinstance(function, SlashCommandParameterAutoCompleter):
            function = function._command_function
        
        command_function = self._command
        if command_function is None:
            
            auto_completer = SlashCommandParameterAutoCompleter(
                function,
                parameter_names,
                APPLICATION_COMMAND_DEEPNESS,
                self,
            )
            
            # If it is none, try to resolve the parameters in sub commands.
            auto_completers = self._auto_completers
            if (auto_completers is None):
                auto_completers = []
                self._auto_completers = auto_completers
            
            auto_completers.append(auto_completer)
            
            sub_commands = self._sub_commands
            if (sub_commands is not None):
                for sub_command in sub_commands.values():
                    sub_command._try_resolve_auto_completer(auto_completer)
        
        else:
            auto_completer = command_function._add_autocomplete_function(parameter_names, function)
        
        _reset_application_command_schema(self)
        
        return auto_completer
    
    
    def _get_self_reference(self):
        """
        Gets a weak reference to the ``SlashCommand``.
        
        Returns
        -------
        self_reference : ``WeakReferer`` to ``SlashCommand``
        """
        self_reference = self._self_reference
        if self_reference is None:
            self_reference = WeakReferer(self)
            self._self_reference = self_reference
        
        return self_reference
    
    
    def _try_resolve_auto_completer(self, auto_completer):
        """
        Tries to register auto completer to the slasher application command.
        
        Parameters
        ----------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The auto completer.
        
        Returns
        -------
        resolved : `int`
            The amount of parameters resolved.
        """
        resolved = 0
        
        command = self._command
        if command is None:
            sub_commands = self._sub_commands
            if (sub_commands is not None):
                for sub_command in sub_commands.values():
                    resolved += sub_command._try_resolve_auto_completer(auto_completer)
        else:
            resolved += command._try_resolve_auto_completer(auto_completer)
        
        if resolved:
            _reset_application_command_schema(self)
        
        return resolved



__all__ = ()

from ...field_parsers import entity_id_parser_factory, force_string_parser_factory, nullable_string_parser_factory
from ...field_putters import entity_id_putter_factory, force_string_putter_factory, nullable_string_putter_factory
from ...field_validators import (
    entity_id_validator_factory, force_string_validator_factory, nullable_string_validator_factory
)

from .constants import CONTENT_LENGTH_MAX, CONTENT_LENGTH_MIN, NAME_LENGTH_MAX, NAME_LENGTH_MIN

# content

parse_content = nullable_string_parser_factory('content')
put_content_into = nullable_string_putter_factory('content')
validate_content = nullable_string_validator_factory('content', CONTENT_LENGTH_MIN, CONTENT_LENGTH_MAX)

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('team_id')

# name

parse_name = force_string_parser_factory('name')
put_name_into = force_string_putter_factory('name')
validate_name = force_string_validator_factory('name', NAME_LENGTH_MIN, NAME_LENGTH_MAX)



from .kokoro_sqlalchemy import *

__all__ = kokoro_sqlalchemy.__all__

from .. import register_library_extension
register_library_extension('HuyaneMatsu.kokoro_sqlalchemy')
del register_library_extension



from .assets import *
from .base import *
from .party import *
from .secrets import *
from .timestamps import *

__all__ = (
    *assets.__all__,
    *base.__all__,
    *party.__all__,
    *secrets.__all__,
    *timestamps.__all__,
)



__all__ = ('Channel',)

import warnings
from collections import deque
from re import I as re_ignore_case, escape as re_escape, match as re_match, search as re_search

from scarletio import LOOP_TIME, copy_docs, export, include

from ....env import MESSAGE_CACHE_SIZE

from ...bases import DiscordEntity
from ...core import CHANNELS, GUILDS
from ...core import MESSAGES
from ...permission.permission import PERMISSION_STAGE_MODERATOR
from ...preconverters import preconvert_preinstanced_type, preconvert_snowflake
from ...user import ZEROUSER, create_partial_user_from_id
from ...utils import DATETIME_FORMAT_CODE

from ..channel_metadata import ChannelMetadataBase, ChannelMetadataGuildMainBase
from ..forum_tag import create_forum_tag_from_id
from ..message_history import MessageHistory, MessageHistoryCollector, message_relative_index

from .preinstanced import ChannelType
from .flags import (
    CHANNEL_TYPE_MASK_CONNECTABLE, CHANNEL_TYPE_MASK_GUILD, CHANNEL_TYPE_MASK_GUILD_SORTABLE,
    CHANNEL_TYPE_MASK_GUILD_SYSTEM, CHANNEL_TYPE_MASK_INVITABLE, CHANNEL_TYPE_MASK_PRIVATE, CHANNEL_TYPE_MASK_TEXTUAL,
    CHANNEL_TYPE_MASK_THREAD, CHANNEL_TYPE_MASK_THREADABLE
)


create_partial_channel_from_id = include('create_partial_channel_from_id')
Message = include('Message')

CHANNEL_TYPE_MASK_GUILD_TEXTUAL = CHANNEL_TYPE_MASK_GUILD | CHANNEL_TYPE_MASK_TEXTUAL
CHANNEL_TYPE_MASK_GUILD_CONNECTABLE = CHANNEL_TYPE_MASK_GUILD | CHANNEL_TYPE_MASK_CONNECTABLE


@export
class Channel(DiscordEntity, immortal = True):
    """
    Represents a Discord channel.
    
    Attributes
    ----------
    id : `int`
        Unique identifier of the channel.
    _message_history : `None`, ``MessageHistory``
        The channel's message history if any.
    guild_id : `int`
        The channel's guild's identifier. Defaults to `0`.
    metadata : ``ChannelMetadataBase``
        The channel's metadata storing it's type specific information.
    type : ``ChannelType``
        The channel's type.
    """
    __slots__ = ('_message_history', 'guild_id', 'metadata', 'type')
    
    def __new__(cls, channel_type = None, **keyword_parameters):
        """
        Creates a partial channel with the given parameters.
        
        Parameters
        ----------
        channel_type : `None`, `int`, ``ChanelType`` = `None`, Optional (Keyword only)
            The channel's type.
        
        **keyword_parameters : Keyword parameters
            Additional predefined attributes for the channel.
        
        Other Parameters
        ----------------
        archived : `bool`, Optional (Keyword only)
            Whether the (thread) channel is archived.
        
        archived_at : `None`, `datetime`, Optional (Keyword only)
            When the thread's archive status was last changed.
        
        applied_tag_ids : `None`, `tuple` of (`int`, ``ForumTag``), Optional (Keyword only)
             The tags' identifier which have been applied to the thread. Applicable for threads of a forum.
        
        auto_archive_after: `int`, Optional (Keyword only)
            The channel's ``.auto_archive_after``.
        
        available_tags : `None`, `tuple` of ``ForumTag``, Optional (Keyword only)
            The available tags to assign to the child-thread channels.
        
        created_at : `None`, `datetime`, Optional (Keyword only)
            When the channel was created.
        
        bitrate : `int`, Optional (Keyword only)
            The bitrate (in bits) of the voice channel.
        
        created_at : `None`, `datetime`, Optional (Keyword only)
            When the (thread) channel was created.
        
        default_forum_layout : ``ForumLayout``, `int`, Optional (Keyword only)
            The default layout used to display threads of the forum.
        
        default_sort_order : ``SortOrder``, `int`, Optional (Keyword only)
            The default thread ordering of the forum.
        
        default_thread_auto_archive_after : `int`, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        
        default_thread_reaction : `None`, ``Emoji``, Optional (Keyword only)
            The emoji to show in the add reaction button on a thread of the forum channel.
        
        default_thread_slowmode : `int`, Optional (Keyword only)
            The default slowmode applied to the channel's threads.
        
        flags : `int`, ``ChannelFlag``, Optional (Keyword only)
            The channel's flags.
        
        icon : `None`, ``Icon``, `str`, Optional (Keyword only)
            The channel's icon.
            
            > Mutually exclusive with `icon_type` and `icon_hash` parameters.
        
        icon_type : ``IconType``, Optional (Keyword only)
            The channel's icon's type.
            
            > Mutually exclusive with the `icon` parameter.
        
        icon_hash : `int`, Optional (Keyword only)
            The channel's icon's hash.
            
            > Mutually exclusive with the `icon` parameter.
        
        invitable : `bool`, Optional (Keyword only)
            Whether non-moderators can invite other non-moderators to the threads.
        
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        nsfw : `int`, Optional (Keyword only)
            Whether the channel is marked as non safe for work.
        
        open : `bool`, Optional (Keyword only)
            Whether the thread channel is open.
        
        parent_id : `None`, `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        
        permission_overwrites : `None`, list` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        region : `None`, ``VoiceRegion``, `str`, Optional (Keyword only)
            The channel's voice region.
        
        slowmode : `int`, Optional (Keyword only)
            The channel's slowmode.
        
        topic : `None`, `str`, Optional (Keyword only)
            The channel's topic.
        
        user_limit : `int`, Optional (Keyword only)
            The maximal amount of users, who can join the voice channel, or `0` if unlimited.
            
        users : `iterable` of (`int`, ``ClientUserBase``), Optional (Keyword only)
            The users in the channel.
            
        video_quality_mode : ``VideoQualityMode``, Optional (Keyword only)
            The video quality of the voice channel.
        
        Returns
        -------
        channel : ``Channel``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        
        if channel_type is None:
            channel_type = ChannelType.unknown
        else:
            channel_type = preconvert_preinstanced_type(channel_type, 'channel_type', ChannelType)
        
        metadata = channel_type.metadata_type.precreate(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Unused or unsettable attributes: {keyword_parameters!r}.'
            )
        
        self = cls._create_empty(0, channel_type, 0)
        self.metadata = metadata
        return self
    
    
    @classmethod
    def from_data(cls, data, client=None, guild_id=0):
        """
        Creates a new channel from the channel data received from Discord. If the channel already exists and if it
        is partial, then updates it.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data receive from Discord.
        client : `None`, ``Client`` = `None`, Optional
            The client, who received the channel's data, if any.
        guild_id : `int` = `0`, Optional
            The guild's identifier of the channel.
        
        Returns
        -------
        self : ``Channel``
        
        Raises
        -------
        RuntimeError
            The respective channel type cannot be instanced.
        """
        channel_id = int(data['id'])
        
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            channel_type = ChannelType.get(data['type'])
            metadata = channel_type.metadata_type.from_data(data)
            
            self = object.__new__(cls)
            self._message_history = None
            self.id = channel_id
            self.guild_id = guild_id
            self.metadata = metadata
            self.type = channel_type
            
            metadata._created(self, client)
            CHANNELS[channel_id] = self
            
        else:
            if self.partial:
                self.guild_id = guild_id
                self._message_history = None
                
                type_ = ChannelType.get(data['type'])
                metadata = type_.metadata_type.from_data(data)
                self.metadata = metadata
                self.type = type_
                metadata._created(self, client)
                
            else:
        
                type_ = ChannelType.get(data['type'])
                if self.type is not type_:
                    metadata = type_.metadata_type.from_data(data)
                    self.metadata = metadata
                    self.type = type_
                    metadata._created(self, client)
        
        return self
    
    
    @classmethod
    def _create_private_data_less(cls, channel_id):
        """
        Creates a private channel from a channel id. Might be called by events, when a message's channel is not found
        and it is a private channel.
        
        Parameters
        ----------
        channel_id : `int`
            The channel's respective id.
        
        Returns
        -------
        self : ``Channel``
        """
        self = cls._create_empty(channel_id, ChannelType.private, 0)
        CHANNELS[channel_id] = self
        return self
    
    
    def _finish_private_data_less(self, client, user):
        """
        Finishes the initialization of the channel after a ``._create_private_data_less`` call.
        
        Parameters
        ----------
        client : ``Client``
            The client recipient of the channel.
        user : ``ClientUserBase``
            The other recipient of the channel.
        """
        metadata = self.metadata
        metadata.users.append(user)
        metadata._created(self, client)
    
    
    def __repr__(self):
        """Returns the representation of the channel."""
        repr_parts = ['<', self.__class__.__name__]
        
        channel_id = self.id
        if channel_id:
            repr_parts.append(' id=')
            repr_parts.append(repr(channel_id))
            repr_parts.append(',')
        
        metadata = self.metadata
        repr_parts.append(' name=')
        repr_parts.append(repr(metadata._get_processed_name()))
        
        type_ = self.type
        repr_parts.append(' type=')
        repr_parts.append(type_.name)
        repr_parts.append('~')
        repr_parts.append(repr(type_.value))
        
        if self.partial:
            repr_parts.append(' (partial)')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the channel's hash value."""
        channel_id = self.id
        if channel_id:
            return channel_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Calculates the channel's hash based on their fields.
        
        This method is called by ``.__hash__`` if the channel has no ``.channel_id`` set.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # guild_id
        hash_value ^= self.guild_id
        
        # metadata
        hash_value ^= hash(self.metadata)
        
        # type
        hash_value ^= self.type.value
        
        return hash_value
    
    
    def __format__(self, code):
        """
        Formats the channel in a format string.
        
        Parameters
        ----------
        code : `str`
            The option on based the result will be formatted.
        
        Returns
        -------
        channel : `str`
        
        Raises
        ------
        ValueError
            Unknown format code.
        
        Examples
        --------
        ```py
        >>> from hata import Channel, now_as_id
        >>> channel = Channel.precreate(now_as_id(), name='GENERAL')
        >>> channel
        <Channel id=710506058560307200, name='GENERAL'>
        >>> # no code stands for `channel.name`.
        >>> f'{channel}'
        'GENERAL'
        >>> # 'd' stands for display name.
        >>> f'{channel:d}'
        'general'
        >>> # 'm' stands for mention.
        >>> f'{channel:m}'
        '<#710506058560307200>'
        >>> # 'c' stands for created at.
        >>> f'{channel:c}'
        '2020.05.14-14:57:24'
        ```
        """
        if not code:
            return self.metadata._get_processed_name()
        
        if code == 'm':
            return self.mention
        
        if code == 'd':
            return self.display_name
        
        if code == 'c':
            return format(self.created_at, DATETIME_FORMAT_CODE)
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"c"!r}, {"d"!r}, {"m"!r}.'
        )
    
    
    @property
    def display_name(self):
        """
        The channel's display name.
        
        Returns
        -------
        display_name : `str`
        """
        return self.metadata._get_display_name()
    
    
    @property
    def mention(self):
        """
        The channel's mention.
        
        Returns
        -------
        mention : `str`
        """
        return f'<#{self.id}>'
    
    
    @property
    def partial(self):
        """
        Whether this channel is partial.
        
        A channel is partial if non of the running clients can see it.
        
        Returns
        -------
        is_partial : `bool`
        """
        if self.id == 0:
            return True
        
        return (not self.clients)
    

    @property
    def clients(self):
        """
        The clients, who can access this channel.
        
        Returns
        -------
        clients : `list` of ``Client``
        """
        return self.metadata._get_clients(self)
    
    
    def get_user(self, name, default = None):
        """
        Tries to find the a user with the given name at the channel. Returns the first matched one.
        
        Parameters
        ----------
        name : `str`
            The name to search for.
        default : `Any` = `None`, Optional
            The value what is returned when no user was found. Defaults to `None`.
        
        Returns
        -------
        user : ``ClientUserBase``, `default`
        """
        return self.metadata._get_user(self, name, default)
    
    
    def get_user_like(self, name, default = None):
        """
        Searches a user, who's name or nick starts with the given string and returns the first find.
        
        Parameters
        ----------
        name : `str`
            The name to search for.
        default : `Any` = `None`, Optional
            The value what is returned when no user was found. Defaults to `None`.
        
        Returns
        -------
        user : ``ClientUserBase``, `default`
        """
        return self.metadata._get_user_like(self, name, default)
    
    
    def get_users_like(self, name):
        """
        Searches the users, who's name or nick starts with the given string.
        
        Parameters
        ----------
        name : `str`
            The name to search for.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        return self.metadata._get_users_like(self, name)
    
    
    @property
    def users(self):
        """
        The users who can see this channel.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        return self.metadata._get_users(self)
    
    
    def iter_users(self):
        """
        Iterates over the users who can see the channel.
        
        This method is a generator.
        
        Yields
        ------
        user : ``ClientUserBase``
        """
        yield from self.metadata._iter_users(self)
    
    
    def has_name_like(self, name):
        """
        Returns whether the channel's name is like the given string.
        
        Parameters
        ----------
        name : `str`
            The name of the channel.
        
        Returns
        -------
        has_name_like : `bool`
        """
        if name.startswith('#'):
            name = name[1:]
        
            match_start = False
        else:
            match_start = True
        
        target_name_length = len(name)
        if (target_name_length < 2) or (target_name_length > 100):
            return False
        
        if match_start:
            matching_function = re_search
        else:
            matching_function = re_match
        
        if matching_function(re_escape(name), self.name, re_ignore_case) is None:
            return False
        
        return True

    
    def permissions_for(self, user):
        """
        Returns the permissions for the given user at the channel.
        
        Parameters
        ----------
        user : ``UserBase``
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
            
        
        See Also
        --------
        ``.cached_permissions_for`` : Cached permission calculator.
        """
        return self.metadata._get_permissions_for(self, user)
    
    
    def cached_permissions_for(self, user):
        """
        Returns the permissions for the given user at the channel. If the user's permissions are not cached, calculates
        and stores them first.
        
        Parameters
        ----------
        user : ``UserBase``
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        """
        return self.metadata._get_cached_permissions_for(self, user)
    
    
    def permissions_for_roles(self, *roles):
        """
        Returns the channel permissions of an imaginary user who would have the listed roles.
        
        Parameters
        ----------
        *roles : ``Role``
            The roles to calculate final permissions from.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        """
        return self.metadata._get_permissions_for_roles(self, roles)
    
    
    @property
    def guild(self):
        """
        Returns the channel's guild. At the case of private channels this is always `None`.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    def iter_threads(self):
        """
        Iterates over the channel's threads.
        
        This method is an iterable generator.
        
        Yields
        ------
        channel : ``Channel``
        """
        guild = self.guild
        if guild is None:
            return
        
        channel_id = self.id
        
        for thread in guild.threads.values():
            if thread.parent_id == channel_id:
                yield thread
    
    
    @property
    def threads(self):
        """
        Returns the channel's threads.
        
        Returns
        -------
        channels : `list` of ``Channel``
        """
        return [*self.iter_threads()]
    
    
    def _update_attributes(self, data):
        """
        Updates the channel with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        """
        type_ = ChannelType.get(data['type'])
        metadata_type = type_.metadata_type
        
        metadata = self.metadata
        
        if metadata_type is type(metadata):
            metadata._update_attributes(data)
        else:
            self.metadata = metadata_type.from_data(data)
        
        self.type = type_
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the channel and returns it's overwritten attributes as a `dict` with a `attribute-name` - `old-value`
        relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
            
            Might contain the following items:
            
            +---------------------------------------+-----------------------------------------------------------+
            | Keys                                  | Values                                                    |
            +=======================================+===========================================================+
            | applied_tag_ids                       | `None`, `tuple` of `int`                                  |
            +---------------------------------------+-----------------------------------------------------------+
            | archived                              | `bool`                                                    |
            +---------------------------------------+-----------------------------------------------------------+
            | archived_at                           | `None`, `datetime`                                        |
            +---------------------------------------+-----------------------------------------------------------+
            | auto_archive_after                    | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | available_tags                        | `None`, `tuple` of ``ForumTag``                           |
            +---------------------------------------+-----------------------------------------------------------+
            | bitrate                               | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | default_forum_layout                  | ``ForumLayout``                                           |
            +---------------------------------------+-----------------------------------------------------------+
            | default_sort_order                    | ``SortOrder``                                             |
            +---------------------------------------+-----------------------------------------------------------+
            | default_thread_auto_archive_after     | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | default_thread_reaction               | `None`, ``Emoji``                                         |
            +---------------------------------------+-----------------------------------------------------------+
            | default_thread_slowmode               | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | flags                                 | ``ChannelFlag``                                           |
            +---------------------------------------+-----------------------------------------------------------+
            | icon                                  | ``Icon``                                                  |
            +---------------------------------------+-----------------------------------------------------------+
            | invitable                             | `bool`                                                    |
            +---------------------------------------+-----------------------------------------------------------+
            | metadata                              | ``ChannelMetadataBase``                                   |
            +---------------------------------------+-----------------------------------------------------------+
            | name                                  | `str`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | nsfw                                  | `bool`                                                    |
            +---------------------------------------+-----------------------------------------------------------+
            | open                                  | `bool`                                                    |
            +---------------------------------------+-----------------------------------------------------------+
            | owner_id                              | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | parent_id                             | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | permission_overwrites                 | `None`, `dict` of (`int`, ``PermissionOverwrite``) items  |
            +---------------------------------------+-----------------------------------------------------------+
            | position                              | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | region                                | `None`, ``VoiceRegion``                                   |
            +---------------------------------------+-----------------------------------------------------------+
            | slowmode                              | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | topic                                 | `None`, `str`                                             |
            +---------------------------------------+-----------------------------------------------------------+
            | type                                  | ``ChannelType``                                           |
            +---------------------------------------+-----------------------------------------------------------+
            | user_limit                            | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | video_quality_mode                    | ``VideoQualityMode``                                      |
            +---------------------------------------+-----------------------------------------------------------+
        """
        type_ = ChannelType.get(data['type'])
        metadata_type = type_.metadata_type
        
        metadata = self.metadata
        
        if metadata_type is type(metadata):
            old_attributes = metadata._difference_update_attributes(data)
            
        else:
            old_attributes = {
                'metadata': metadata,
            }
            
            self.metadata = metadata_type.from_data(data)
        
        if (type_ is not self.type):
            old_attributes['type'] = self.type
            self.type = type_
        
        return old_attributes
    
    
    def _delete(self, client):
        """
        Called when the channel is deleted.
        
        Removes the channel's references.
        
        Parameters
        ----------
        client : `None`, ``Client``
            The parent client entity.
        """
        self.metadata._delete(self, client)
    
    
    def _iter_delete(self, client):
        """
        Called when a channel is deleted. Not like ``._delete`` this will apply deletion to all related channel as well,
        technically calling ``._delete`` on all.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        client : `None`, ``Client``
            The parent client entity.
        
        Yields
        ------
        channel : ``Channel``
        """
        yield from self.metadata._iter_delete(self, client)
    
    
    @classmethod
    def _from_partial_data(cls, data, channel_id, guild_id):
        """
        Creates a channel from partial data. Called by ``create_partial_channel_from_data`` when a new
        partial channel is needed to be created.
        
        Parameters
        ----------
        data : `None`, `dict` of (`str`, `Any`) items
            Partial channel data.
        channel_id : `int`
            The channel's id.
        guild_id : `int`
            The channel's guild's identifier if applicable.
        
        Returns
        -------
        channel : ``Channel``
        """
        type_ = ChannelType.get(data.get('type', -1))
        metadata = type_.metadata_type._from_partial_data(data)
        
        self = object.__new__(cls)
        self._message_history = None
        self.id = channel_id
        self.guild_id = guild_id
        self.metadata = metadata
        self.type = type_
        
        return self
    
    
    @classmethod
    def _create_empty(cls, channel_id, channel_type, guild_id):
        """
        Creates a partial channel from the given parameters.
        
        Parameters
        ----------
        channel_id : `int`
            The channel's identifier.
        channel_type : ``ChannelType``
            The channel's type.
        guild_id : `int`
            A partial guild's identifier for the created channel.
        
        Returns
        -------
        channel : ``Channel``
            The created partial channel.
        """
        self = object.__new__(cls)
        self._message_history = None
        self.id = channel_id
        self.guild_id = guild_id
        self.metadata = channel_type.metadata_type._create_empty()
        self.type = channel_type
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the channel to json serializable representation dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether we want to include identifiers as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        data = self.metadata.to_data(defaults = defaults, include_internals = include_internals)
        
        # id
        if include_internals:
            data['id'] = str(self.id)
        
        # type
        data['type'] = self.type.value
        
        # guild_id
        if include_internals:
            guild_id = self.guild_id
            if guild_id:
                guild_id = str(guild_id)
            else:
                guild_id = None
            
            data['guild_id'] = guild_id
        
        return data
    
    
    @property
    def created_at(self):
        """
        Returns when the channel was created.
        
        Returns
        -------
        created_at : `datetime`
        """
        return self.metadata._get_created_at(self)
    
    
    @property
    def order_group(self):
        """
        Returns the channel's order group.
        
        Returns
        -------
        order_group : `int`
        """
        return self.metadata.order_group
    
    
    # for sorting channels
    def __gt__(self, other):
        """Returns whether this channel's is greater than the other's."""
        if type(self) is not type(other):
            return NotImplemented
        
        
        self_metadata = self.metadata
        other_metadata = other.metadata
        
        if (
            isinstance(self_metadata, ChannelMetadataGuildMainBase) and
            isinstance(other_metadata, ChannelMetadataGuildMainBase)
        ):
            self_order_group = self_metadata.order_group
            other_order_group = other_metadata.order_group
            
            if self_order_group > other_order_group:
                return True
            
            if self_order_group != self_order_group:
                return False
                
            self_position = self_metadata.position
            other_position = other_metadata.position
            
            if self_position > other_position:
                return True
            
            if self_position != other_position:
                return False
        
        
        return self.id > other.id
    
    
    def __ge__(self, other):
        """Returns whether this channel's is greater or equal than the other's."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_metadata = self.metadata
        other_metadata = other.metadata
        
        if (
            isinstance(self_metadata, ChannelMetadataGuildMainBase) and
            isinstance(other_metadata, ChannelMetadataGuildMainBase)
        ):
            self_order_group = self_metadata.order_group
            other_order_group = other_metadata.order_group
            
            if self_order_group > other_order_group:
                return True
            
            if self_order_group != self_order_group:
                return False
                
            self_position = self_metadata.position
            other_position = other_metadata.position
            
            if self_position > other_position:
                return True
            
            if self_position != other_position:
                return False
        
        
        return self.id >= other.id
        
    
    def __eq__(self, other):
        """Returns whether this channel's is equal to the other's."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_time(other)
        
    
    def __ne__(self,other):
        """Returns whether this channel's is not equal to the other's."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_time(other)
    
    
    def _is_equal_same_time(self, other):
        """
        Returns whether the channel is equal to the other one.
        
        Parameters
        ----------
        other : ``Channel``
            The other channel entity.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            if self_id == other_id:
                return True
            
            return False
        
        # guild_id
        # Internal field -> ignore
        
        # metadata
        if self.metadata != other.metadata:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        return True
    
    
    def __le__(self, other):
        """Returns whether this channel's is less or equal than the other's."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_metadata = self.metadata
        other_metadata = other.metadata
        
        if (
            isinstance(self_metadata, ChannelMetadataGuildMainBase) and
            isinstance(other_metadata, ChannelMetadataGuildMainBase)
        ):
            self_order_group = self_metadata.order_group
            other_order_group = other_metadata.order_group
            
            if self_order_group < other_order_group:
                return True
            
            if self_order_group != self_order_group:
                return False
            
            self_position = self_metadata.position
            other_position = other_metadata.position
            
            if self_position < other_position:
                return True
            
            if self_position != other_position:
                return False
        
        
        return self.id <= other.id
        
    
    def __lt__(self, other):
        """Returns whether this channel's is less than the other's."""
        if type(self) is not type(other):
            return NotImplemented
        
        self_metadata = self.metadata
        other_metadata = other.metadata
        
        if (
            isinstance(self_metadata, ChannelMetadataGuildMainBase) and
            isinstance(other_metadata, ChannelMetadataGuildMainBase)
        ):
            self_order_group = self_metadata.order_group
            other_order_group = other_metadata.order_group
            
            if self_order_group < other_order_group:
                return True
            
            if self_order_group != self_order_group:
                return False
                
            self_position = self_metadata.position
            other_position = other_metadata.position
            
            if self_position < other_position:
                return True
            
            if self_position != other_position:
                return False
        
        
        return self.id < other.id
    
    # Fields
    
    @property
    @copy_docs(ChannelMetadataBase.applied_tag_ids)
    def applied_tag_ids(self):
        return self.metadata.applied_tag_ids
    
    
    @property
    @copy_docs(ChannelMetadataBase.archived)
    def archived(self):
        return self.metadata.archived
    
    
    @property
    @copy_docs(ChannelMetadataBase.archived_at)
    def archived_at(self):
        return self.metadata.archived_at
    
    
    @property
    @copy_docs(ChannelMetadataBase.auto_archive_after)
    def auto_archive_after(self):
        return self.metadata.auto_archive_after
    
    
    @property
    @copy_docs(ChannelMetadataBase.available_tags)
    def available_tags(self):
        return self.metadata.available_tags
    
    
    @property
    @copy_docs(ChannelMetadataBase.bitrate)
    def bitrate(self):
        return self.metadata.bitrate
    
    
    @property
    @copy_docs(ChannelMetadataBase.default_forum_layout)
    def default_forum_layout(self):
        return self.metadata.default_forum_layout
    
    
    @property
    @copy_docs(ChannelMetadataBase.default_sort_order)
    def default_sort_order(self):
        return self.metadata.default_sort_order
    
    
    @property
    @copy_docs(ChannelMetadataBase.default_thread_auto_archive_after)
    def default_thread_auto_archive_after(self):
        return self.metadata.default_thread_auto_archive_after
    
    
    @property
    @copy_docs(ChannelMetadataBase.default_thread_reaction)
    def default_thread_reaction(self):
        return self.metadata.default_thread_reaction
    
    
    @property
    @copy_docs(ChannelMetadataBase.default_thread_slowmode)
    def default_thread_slowmode(self):
        return self.metadata.default_thread_slowmode
    
    
    @property
    @copy_docs(ChannelMetadataBase.flags)
    def flags(self):
        return self.metadata.flags
    
    
    @property
    @copy_docs(ChannelMetadataBase.icon)
    def icon(self):
        return self.metadata.icon
    
    
    @property
    @copy_docs(ChannelMetadataBase.invitable)
    def invitable(self):
        return self.metadata.invitable
    
    
    @property
    @copy_docs(ChannelMetadataBase.name)
    def name(self):
        return self.metadata.name
    
    
    @property
    @copy_docs(ChannelMetadataBase.nsfw)
    def nsfw(self):
        return self.metadata.nsfw
    
    
    @property
    @copy_docs(ChannelMetadataBase.open)
    def open(self):
        return self.metadata.open
    
    
    @property
    @copy_docs(ChannelMetadataBase.owner_id)
    def owner_id(self):
        return self.metadata.owner_id
    
    
    @property
    @copy_docs(ChannelMetadataBase.parent_id)
    def parent_id(self):
        return self.metadata.parent_id
    
    
    @property
    @copy_docs(ChannelMetadataBase.permission_overwrites)
    def permission_overwrites(self):
        return self.metadata.permission_overwrites
    
    
    @property
    @copy_docs(ChannelMetadataBase.position)
    def position(self):
        return self.metadata.position
    
    
    @property
    @copy_docs(ChannelMetadataBase.region)
    def region(self):
        return self.metadata.region
    
    
    @property
    @copy_docs(ChannelMetadataBase.slowmode)
    def slowmode(self):
        return self.metadata.slowmode
    
    
    @property
    @copy_docs(ChannelMetadataBase.topic)
    def topic(self):
        return self.metadata.topic
    
    
    @property
    @copy_docs(ChannelMetadataBase.thread_users)
    def thread_users(self):
        return self.metadata.thread_users
    
    
    @thread_users.setter
    def thread_users(self, thread_users):
        self.metadata.thread_users = thread_users
    
    
    @property
    @copy_docs(ChannelMetadataBase.user_limit)
    def user_limit(self):
        return self.metadata.user_limit
    
    
    @property
    @copy_docs(ChannelMetadataBase.video_quality_mode)
    def video_quality_mode(self):
        return self.metadata.video_quality_mode
    
    # Utility
    
    @property
    def parent(self):
        """
        Returns the channel's parent.
        
        If the channel has no parent, or if not applicable for the specific channel type returns `None`.
        
        Returns
        -------
        parent : `None`, ``Channel``
        """
        parent_id = self.metadata.parent_id
        if parent_id:
            return create_partial_channel_from_id(parent_id, ChannelType.unknown, self.guild_id)
    
    
    @property
    def applied_tags(self):
        """
        Returns the applied tags to the thread.
        
        Returns
        -------
        applied_tags : `None`, `tuple` of ``ForumTag``
        """
        applied_tag_ids = self.applied_tag_ids
        if (applied_tag_ids is None):
            return None
        
        return tuple(create_forum_tag_from_id(forum_tag_id) for forum_tag_id in applied_tag_ids)
    
    
    def iter_applied_tag_ids(self):
        """
        Iterates over the applied tag identifiers of the channel.
        
        This method is an iterable generator.
        
        Yields
        ------
        applied_tag_id : `int`
        """
        applied_tag_ids = self.applied_tag_ids
        if (applied_tag_ids is not None):
            yield from applied_tag_ids
    
    
    def iter_applied_tags(self):
        """
        Iterates over the applied tags of the channels.
        
        This method is an iterable generator.
        
        Yields
        ------
        applied_tag : ``ForumTag``
        """
        applied_tag_ids = self.applied_tag_ids
        if (applied_tag_ids is not None):
            for forum_tag_id in applied_tag_ids:
                yield create_forum_tag_from_id(forum_tag_id)
    
    
    def iter_available_tags(self):
        """
        Iterates over the available tags of the the forum channel.
        
        This method is an iterable generator.
        
        Yields
        ------
        available_tag : ``ForumTag``
        """
        available_tags = self.available_tags
        if (available_tags is not None):
            yield from available_tags
    
    
    @property
    def owner(self):
        """
        Returns the group channel's owner.
        
        Returns
        -------
        owner : ``ClientUserBase``
            Defaults to `ZEROUSER`.
        """
        owner_id = self.owner_id
        if owner_id:
            owner = create_partial_user_from_id(owner_id)
        else:
            owner = ZEROUSER
        return owner
    
    
    @property
    def channel_list(self):
        """
        Returns the channels of the category in a list in their display order.
        
        Returns
        -------
        channels : `list` of ``Channel``
        """
        channels = []
        
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                channel_id = self.id
                
                for channel in guild.channels.values():
                    if channel.parent_id == channel_id:
                        channels.append(channel)
                
                channels.sort()
        
        return channels
    
    
    @property
    def voice_users(self):
        """
        Returns a list of the users who are in the voice channel.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        users = []
        
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                channel_id = self.id
                
                for voice_state in guild.voice_states.values():
                    if voice_state.channel_id == channel_id:
                        users.append(voice_state.user)
        
        return users
    
    
    @property
    def audience(self):
        """
        Returns the audience in the stage channel.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        users = []
        
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                channel_id = self.id
                
                for voice_states in guild.voice_states.values():
                    if (voice_states.channel_id == channel_id) and voice_states.is_speaker:
                        users.append(voice_states.user)
        
        return users
    
    
    @property
    def speakers(self):
        """
        Returns the speakers in the stage channel.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        users = []
        
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                channel_id = self.id
                
                for voice_states in guild.voice_states.values():
                    if (voice_states.channel_id == channel_id) and (not voice_states.is_speaker):
                        users.append(voice_states.user)
        
        return users
    
    
    @property
    def moderators(self):
        """
        Returns the moderators in the stage channel.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        users = []
        
        guild_id = self.guild_id
        if guild_id:
            try:
                guild = GUILDS[guild_id]
            except KeyError:
                pass
            else:
                channel_id = self.id
                
                for voice_states in guild.voice_states.values():
                    if (voice_states.channel_id == channel_id):
                        user = voice_states.user
                        if self.permissions_for(user) >= PERMISSION_STAGE_MODERATOR:
                            users.append(user)
        
        return users
    
    
    @classmethod
    def precreate(cls, channel_id, *, channel_type = None, guild_id = 0, **keyword_parameters):
        """
        Precreates the channel by creating a partial one with the given parameters. When the channel is loaded
        the precreated channel will be picked up. If an already existing channel would be precreated, returns that
        instead and updates that only, if that is a partial channel.
        
        Parameters
        ----------
        channel_id : `int`, `str`
            The channel's id.
        
        channel_type : `None`, `int`, ``ChanelType`` = `None`, Optional (Keyword only)
            The channel's type.

        guild_id : `int` = `0`, Optional (Keyword only)
            The channel's parent guild's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional predefined attributes for the channel.
        
        Other Parameters
        ----------------
        archived : `bool`, Optional (Keyword only)
            Whether the (thread) channel is archived.
        
        archived_at : `None`, `datetime`, Optional (Keyword only)
            When the thread's archive status was last changed.
        
        applied_tag_ids : `None`, `tuple` of (`int`, ``ForumTag``), Optional (Keyword only)
             The tags' identifier which have been applied to the thread. Applicable for threads of a forum.
        
        auto_archive_after: `int`, Optional (Keyword only)
            The channel's ``.auto_archive_after``.
        
        available_tags : `None`, `tuple` of ``ForumTag``, Optional (Keyword only)
            The available tags to assign to the child-thread channels.
        
        created_at : `None`, `datetime`, Optional (Keyword only)
            When the channel was created.
        
        bitrate : `int`, Optional (Keyword only)
            The channel's ``.bitrate``.
        
        created_at : `None`, `datetime`, Optional (Keyword only)
            When the (thread) channel was created.
        
        default_forum_layout : ``ForumLayout``, `int`, Optional (Keyword only)
            The default layout used to display threads of the forum.
        
        default_sort_order : ``SortOrder``, `int`, Optional (Keyword only)
            The default thread ordering of the forum.
        
        default_thread_auto_archive_after : `int`, Optional (Keyword only)
            The channel's ``.default_thread_auto_archive_after``.
        
        default_thread_slowmode : `int`, Optional (Keyword only)
            The default slowmode applied to the channel's threads.
        
        default_thread_reaction : `None`, ``Emoji``, Optional (Keyword only)
            The emoji to show in the add reaction button on a thread of the forum channel.
        
        flags : `int`, ``ChannelFlag``, Optional (Keyword only)
            The channel's ``.flags``.
        
        icon : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The channel's icon.
            
            > Mutually exclusive with `icon_type` and `icon_hash` parameters.
        
        icon_type : ``IconType``, Optional (Keyword only)
            The channel's icon's type.
            
            > Mutually exclusive with the `icon` parameter.
        
        icon_hash : `int`, Optional (Keyword only)
            The channel's icon's hash.
            
            > Mutually exclusive with the `icon` parameter.
        
        invitable : `bool`, Optional (Keyword only)
            The channel's `..invitable``.
        
        name : `str`, Optional (Keyword only)
            The channel's ``.name``.
        
        nsfw : `int`, Optional (Keyword only)
            The channel's ``.nsfw``.
        
        open : `bool`, Optional (Keyword only)
            The channel's ``.open``.
        
        parent_id : `None`, `int`, ``Channel``, Optional (Keyword only)
            The channel's parent's identifier.
        
        permission_overwrites : `None`, list` of ``PermissionOverwrite``, Optional (Keyword only)
            The channel's permission overwrites.
        
        position : `int`, Optional (Keyword only)
            The channel's position.
        
        region : `None`, ``VoiceRegion``, `str`, Optional (Keyword only)
            The channel's voice region.
        
        slowmode : `int`, Optional (Keyword only)
            The channel's ``.slowmode``.
        
        topic : `None`, `str`, Optional (Keyword only)
            The channel's ``.topic``.
        
        user_limit : `int`, Optional (Keyword only)
            The channel's ``.user_limit``.
            
        users : `iterable` of (`int`, ``ClientUserBase``), Optional (Keyword only)
            The users in the channel.
            
        video_quality_mode : ``VideoQualityMode``, Optional (Keyword only)
            The video quality of the voice channel.
        
        Returns
        -------
        channel : ``Channel``
        
        Raises
        ------
        TypeError
            If any parameter's type is bad or if unexpected parameter is passed.
        ValueError
            If an parameter's type is good, but it's value is unacceptable.
        """
        channel_id = preconvert_snowflake(channel_id, 'channel_id')
        guild_id = preconvert_snowflake(guild_id, 'guild_id')
        
        if channel_type is None:
            channel_type = ChannelType.unknown
        else:
            channel_type = preconvert_preinstanced_type(channel_type, 'channel_type', ChannelType)
        
        metadata = channel_type.metadata_type.precreate(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Unused or unsettable attributes: {keyword_parameters!r}.'
            )
        
        
        try:
            self = CHANNELS[channel_id]
        except KeyError:
            self = cls._create_empty(channel_id, channel_type, guild_id)
            CHANNELS[channel_id] = self
        else:
            if not self.partial:
                return self
        
        self.metadata = metadata
        return self
    
    
    # ---- Messaging ----
    
    @property
    def message_history_reached_end(self):
        message_history = self._message_history
        if message_history is None:
            message_history_reached_end = False
        else:
            message_history_reached_end = message_history.message_history_reached_end
        
        return message_history_reached_end
    
    @message_history_reached_end.setter
    def message_history_reached_end(self, message_history_reached_end):
        message_history = self._message_history
        if message_history is None:
            message_history = MessageHistory(MESSAGE_CACHE_SIZE)
            self._message_history = message_history
        
        message_history.message_history_reached_end = message_history_reached_end
    
    
    @property
    def _message_history_collector(self):
        message_history = self._message_history
        if message_history is None:
            message_history_collector = None
        else:
            message_history_collector = message_history._message_history_collector
        
        return message_history_collector
    
    @_message_history_collector.setter
    def _message_history_collector(self, message_history_collector):
        message_history = self._message_history
        if message_history is None:
            message_history = MessageHistory(MESSAGE_CACHE_SIZE)
            self._message_history = message_history
        
        message_history._message_history_collector = message_history_collector
    
    
    @property
    def _message_keep_limit(self):
        message_history = self._message_history
        if message_history is None:
            message_keep_limit = MESSAGE_CACHE_SIZE
        else:
            message_keep_limit = message_history._message_keep_limit
        
        return message_keep_limit
    
    @_message_keep_limit.setter
    def _message_keep_limit(self, message_keep_limit):
        message_history = self._message_history
        if message_history is None:
            message_history = MessageHistory(MESSAGE_CACHE_SIZE)
            self._message_history = message_history
        
        message_history._message_keep_limit = message_keep_limit
    
    
    @property
    def messages(self):
        message_history = self._message_history
        if message_history is None:
            messages = None
        else:
            messages = message_history.messages
        
        return messages
    
    @messages.setter
    def messages(self, messages):
        message_history = self._message_history
        if message_history is None:
            message_history = MessageHistory(MESSAGE_CACHE_SIZE)
            self._message_history = message_history
        
        message_history.messages = messages
    
    
    @property
    def message_keep_limit(self):
        """
        A property for getting or setting how much message the channel can store before removing the last.
        
        Returns and accepts an `int`.
        """
        message_history = self._message_history
        if (message_history is None):
            message_keep_limit = MESSAGE_CACHE_SIZE
        else:
            message_keep_limit = message_history._message_keep_limit
        
        return message_keep_limit
    
    @message_keep_limit.setter
    def message_keep_limit(self, message_keep_limit):
        if message_keep_limit <= 0:
            self._message_history = None
        
        else:
            message_history = self._message_history
            if (message_history is None):
                self._message_history = MessageHistory(message_keep_limit)
            else:
                message_history._set_message_keep_limit(message_keep_limit)
    
    
    def _create_new_message(self, message_data):
        """
        Creates a new message at the channel. If the message already exists inside of the channel's message history,
        returns that instead.
        
        Parameters
        ----------
        message_data : `dict` of (`str`, `Any`) items
            Message data received from Discord.
        
        Returns
        -------
        message : ``Message``
        """
        from_cache, message = Message._create_message_is_in_cache(message_data)
        if from_cache:
            message._late_init(message_data)
            return message
        
        messages = self._maybe_create_queue()
        message_id = message.id
        
        if (messages is not None):
            if messages and (messages[0].id > message_id):
                index = message_relative_index(messages, message_id)
                max_length = messages.maxlen
                if max_length is None:
                    max_length_reached = False
                else:
                    if max_length == len(messages):
                        max_length_reached = True
                    else:
                        max_length_reached = False
                
                if index == len(messages):
                    if not max_length_reached:
                        messages.append(message)
                else:
                    if max_length_reached:
                        messages.pop()
                        self.message_history_reached_end = False
                    
                    messages.insert(index, message)
            
            else:
                messages_length = len(messages)
                messages.appendleft(message)
                if messages_length != len(messages):
                    self.message_history_reached_end = False
        
        return message
    
    
    def _create_old_message(self, message_data):
        """
        Creates an old message at the channel. If the message already exists inside of the channel's message history,
        returns that instead.
        
        Parameters
        ----------
        message_data : `dict` of (`str`, `Any`) items
            Message data received from Discord.
        
        Returns
        -------
        message : ``Message``
        
        Notes
        -----
        The created message cannot be added to the channel's message history, if it has no more spaces.
        """
        message = Message.from_data(message_data)
        message_id = message.id
        
        messages = self.messages
        if (messages is not None) and messages and (message_id > messages[-1].id):
            index = message_relative_index(messages, message_id)
            if index != len(messages):
                if messages[index].id != message_id:
                    self._maybe_increase_queue_size().insert(index, message)
        else:
            self._maybe_increase_queue_size().append(message)
        
        return message
    
    
    def _create_find_message(self, message_data, chained):
        """
        Tries to find whether the given message's data represents an existing message at the channel. If not, creates
        it. This method also returns whether the message existed at the channel's message history.
        
        Parameters
        ----------
        message_data : `dict` of (`str`, `Any`) items
            The message's data to find or create.
        chained : `bool`
            Whether the created message should be chained to the channel's message history's end, if not found.
        
        Returns
        -------
        message : ``Message``
        found : `bool`
        """
        message_id = int(message_data['id'])
        messages = self.messages
        if (messages is not None):
            index = message_relative_index(messages, message_id)
            if index != len(messages):
                message = messages[index]
                if message.id == message_id:
                    return message, True
        
        message = Message.from_data(message_data)
        
        if chained:
            self._maybe_increase_queue_size().append(message)
        
        return message, False
    
    
    def _add_message_collection_delay(self, delay):
        """
        Sets message collection timeout to the exact given time.
        
        Parameters
        ----------
        delay : `float`
            The time to delay the message collection with.
        """
        message_history_collector = self._message_history_collector
        if (message_history_collector is None):
            self._message_history_collector = MessageHistoryCollector(self, LOOP_TIME() + delay)
        else:
            message_history_collector.add_delay(delay)
    
    
    def _cancel_message_collection(self):
        """
        Cancels the message collector of the channel.
        """
        message_history_collector = self._message_history_collector
        if (message_history_collector is not None):
            self._message_history_collector = None
            message_history_collector.cancel()
    
    
    def _maybe_increase_queue_size(self):
        """
        Increases the queue size of the channel's message history if needed and returns it.
        
        Returns
        -------
        messages : `deque`
        """
        messages = self.messages
        if messages is None:
            # Create unlimited size.
            self.messages = messages = deque()
            self._add_message_collection_delay(110.0)
        else:
            max_length = messages.maxlen
            if (max_length is None):
                # The size is already unlimited
                self._add_message_collection_delay(10.0)
            else:
                # Switch to unlimited if we hit our current limit.
                if len(messages) == max_length:
                    self.messages = messages = deque(messages)
                    self._add_message_collection_delay(110.0)
        
        return messages
    
    
    def _maybe_create_queue(self):
        """
        Gets the channel's messages when creating a new message is created.
        
        Returns
        -------
        messages : `deque`, `None`
        """
        messages = self.messages
        if messages is None:
            message_keep_limit = self._message_keep_limit
            if message_keep_limit == 0:
                if self._message_history_collector is None:
                    messages = None
                else:
                    self.messages = messages = deque(maxlen=None)
            else:
                self.messages = messages = deque(maxlen=message_keep_limit)
        else:
            
            max_length = messages.maxlen
            if (max_length is not None) and (len(messages) == max_length):
                if self._message_history_collector is None:
                    self.message_history_reached_end = False
                else:
                    self.messages = messages = deque(messages, maxlen=None)
        
        return messages
    
    
    def _switch_to_limited(self):
        """
        Switches a channel's `.messages` to limited from unlimited.
        """
        old_messages = self.messages
        if old_messages is None:
            new_messages = None
        else:
            limit = self._message_keep_limit
            if limit == 0:
                new_messages = None
            else:
                new_messages = deque(
                    (old_messages[index] for index in range(min(limit, len(old_messages)))),
                    maxlen = limit,
                )
        
        self.messages = new_messages
        self._cancel_message_collection()
        self.message_history_reached_end = False
    
    
    def _pop_message(self, delete_id):
        """
        Removes the specific message by it's id from the channel's message history and from `MESSAGES` as well.
        
        Parameters
        ----------
        delete_id : `int`
            The message's id to delete from the channel's message history.
        
        Returns
        -------
        message : `None`, ``Message``
        """
        messages = self.messages
        if (messages is not None):
            index = message_relative_index(messages, delete_id)
            if index != len(messages):
                message = messages[index]
                if message.id == delete_id:
                    del messages[index]
                    if (self._message_history_collector is not None):
                        if len(messages) < self._message_keep_limit:
                            self._switch_to_limited()
                    
                    try:
                        del MESSAGES[delete_id]
                    except KeyError:
                        pass
                    
                    message.deleted = True
                    return message
        
        try:
            message = MESSAGES.pop(delete_id)
        except KeyError:
            message = None
        else:
            message.deleted = True
        
        return message
    
    def _pop_multiple(self, delete_ids):
        """
        Removes the given messages from the channel and from `MESSAGES` as well. Returns the found messages.
        
        Parameters
        ----------
        delete_ids : `list` of `int`
            The messages' id to delete from the channel's message history.
        
        Returns
        -------
        found : `list` of ``Message``
            The found and removed messages.
        missed : `list` of `int`
            The identifier of the not found messages.
        """
        found = []
        missed = []
        delete_length = len(delete_ids)
        if not delete_length:
            return found, missed
        
        messages = self.messages
        delete_ids.sort(reverse=True)
        if messages is None:
            messages_length = 0
        else:
            messages_length = len(messages)
        
        if messages is None:
            messages_index = 0
        else:
            messages_index = message_relative_index(messages, delete_ids[0])
        delete_index = 0
        
        while True:
            if delete_index == delete_length:
                break
            
            if messages_index == messages_length:
                while True:
                    delete_id = delete_ids[delete_index]
                    try:
                        message = MESSAGES.pop(delete_id)
                    except KeyError:
                        missed.append(delete_id)
                    else:
                        message.deleted = True
                        found.append(message)
                        
                    delete_index += 1
                    if delete_index == delete_length:
                        break
                    
                    continue
                break
            
            message = messages[messages_index]
            delete_id = delete_ids[delete_index]
            message_id = message.id
            
            if message_id == delete_id:
                del messages[messages_index]
                try:
                    del MESSAGES[delete_id]
                except KeyError:
                    pass
                
                message.deleted = True
                found.append(message)
                
                messages_length -= 1
                delete_index += 1
                continue
            
            if message_id > delete_id:
                messages_index += 1
                continue
            
            delete_index += 1
            
            try:
                message = MESSAGES.pop(delete_id)
            except KeyError:
                missed.append(delete_id)
            else:
                message.deleted = True
                found.append(message)
            
            continue
        
        if (
            (messages is not None) and
            (self._message_history_collector is not None) and
            (len(messages) < self._message_keep_limit)
        ):
            self._switch_to_limited()
        
        return found, missed
    
    # ---- Utility methods ----
    
    def is_in_group_messageable(self):
        """
        Deprecated and will be removed in 2023 January. Please use ``.is_in_group_textual`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.is_in_group_messageable` is deprecated and will be removed in '
                f'2023 January. Please use `.is_in_group_textual` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_in_group_textual()
    
    
    def is_in_group_textual(self):
        """
        Returns whether the channel is messageable.
        
        Returns
        -------
        is_in_group_textual : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_TEXTUAL
    
    
    def is_in_group_guild_messageable(self):
        """
        Deprecated and will be removed in 2023 January. Please use ``.is_in_group_guild_textual`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.is_in_group_guild_messageable` is deprecated and will be removed in '
                f'2023 January. Please use `.is_in_group_guild_textual` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_in_group_guild_textual()
    
    
    def is_in_group_guild_textual(self):
        """
        Returns whether the channel is a guild and messageable one.
        
        Returns
        -------
        is_in_group_guild_textual : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_GUILD_TEXTUAL == CHANNEL_TYPE_MASK_GUILD_TEXTUAL
    
    
    def is_in_group_guild_main_text(self):
        """
        Deprecated and will be removed in 2023 January. Please use ``.is_in_group_guild_system`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.is_in_group_guild_main_text` is deprecated and will be removed in '
                f'2023 January. Please use `.is_in_group_guild_system` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_in_group_guild_system()
    
    
    def is_in_group_guild_system(self):
        """
        Returns whether the channel a guild text like channel.
        
        > Excludes connectable and thread channels.
        
        Returns
        -------
        is_in_group_guild_system : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_GUILD_SYSTEM
    
    
    def is_in_group_connectable(self):
        """
        Returns whether the channel is connectable.
        
        > Includes private connectable channels. Bots cannot connect to those, so you might want to use
        > ``.is_guild_connectable`` instead.
        
        Returns
        -------
        is_in_group_connectable : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_CONNECTABLE
    
    
    def is_in_group_guild_connectable(self):
        """
        Returns whether the channel is a guild connectable channel.
        
        Returns
        -------
        is_in_group_guild_connectable : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_GUILD_CONNECTABLE == CHANNEL_TYPE_MASK_GUILD_CONNECTABLE
    
    
    def is_in_group_private(self):
        """
        Returns whether the channel is a private channel.
        
        Returns
        -------
        is_in_group_private : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_PRIVATE
    
    
    def is_in_group_guild(self):
        """
        Returns whether the channel is a guild channel.
        
        Returns
        -------
        is_in_group_guild : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_GUILD
    
    
    def is_in_group_thread(self):
        """
        Returns whether the channel is a thread.
        
        Returns
        -------
        is_in_group_thread : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_THREAD
    
    
    def is_in_group_can_contain_threads(self):
        """
        Deprecated and will be removed in 2023 January. Please use ``.is_in_group_threadable`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.is_in_group_can_contain_threads` is deprecated and will be removed in '
                f'2023 January. Please use `.is_in_group_threadable` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_in_group_threadable()
    
    
    def is_in_group_threadable(self):
        """
        Returns whether the channel can have threads.
        
        Returns
        -------
        is_in_group_threadable : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_THREADABLE
    
    
    def is_in_group_can_create_invite_to(self):
        """
        Deprecated and will be removed in 2023 January. Please use ``.is_in_group_invitable`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.is_in_group_can_create_invite_to` is deprecated and will be removed in '
                f'2023 January. Please use `.is_in_group_invitable` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_in_group_invitable()
    
    
    def is_in_group_invitable(self):
        """
        Returns whether the channel have invites created to.
        
        Returns
        -------
        is_in_group_invitable : `bool`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_INVITABLE
    
    
    def is_in_group_guild_movable(self):
        """
        Deprecated and will be removed in 2023 January. Please use ``.is_in_group_guild_sortable`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.is_in_group_guild_movable` is deprecated and will be removed in '
                f'2023 January. Please use `.is_in_group_guild_sortable` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return self.is_in_group_guild_sortable()
    
    
    def is_in_group_guild_sortable(self):
        """
        Returns whether the channel is a movable guild channel.
        
        Returns
        -------
        is_in_group_guild_sortable : `int`
        """
        return self.type.flags & CHANNEL_TYPE_MASK_GUILD_SORTABLE
    
    
    def is_guild_text(self):
        """
        Returns whether the channel is a guild text channel.
        
        > Excludes Announcements and thread channels.
        
        Returns
        -------
        is_guild_text : `int`
        """
        return self.type is ChannelType.guild_text
    
    
    def is_private(self):
        """
        Returns whether the channel is a private or direct message (DM) channel.
        
        Returns
        -------
        is_private : `int`
        """
        return self.type is ChannelType.private
        
        
    def is_guild_voice(self):
        """
        Returns whether the guild is a guild voice channel.
        
        > Excludes stage channels.
        
        Returns
        -------
        is_guild_voice : `int`
        """
        return self.type is ChannelType.guild_voice
    
    
    def is_private_group(self):
        """
        Returns whether the channel is a private group channel.
        
        Returns
        -------
        is_private_group : `int`
        """
        return self.type is ChannelType.private_group
    
    
    def is_guild_category(self):
        """
        Returns whether the guild is a guild directory channel.
        
        Returns
        -------
        is_guild_category : `int`
        """
        return self.type is ChannelType.guild_category
    
    
    def is_guild_announcements(self):
        """
        Returns whether the channel is a guild announcements channel.
        
        Returns
        -------
        is_guild_announcements : `int`
        """
        return self.type is ChannelType.guild_announcements
    
    
    def is_guild_store(self):
        """
        Returns whether the channel is a guild store channels.
        
        > Store channels are deprecated & removed from Discord.
        
        Returns
        -------
        is_guild_store : `int`
        """
        return self.type is ChannelType.guild_store
    
    
    def is_thread(self):
        """
        Returns whether the channel is a thread channel.
        
        > This thread channel type never made into Discord.
        
        Returns
        -------
        is_thread : `bool`
        """
        return self.type is ChannelType.thread
    
    
    def is_guild_thread_announcements(self):
        """
        Returns whether the channel is a guild announcements thread.
        
        Returns
        -------
        is_guild_thread_announcements : `bool`
        """
        return self.type is ChannelType.guild_thread_announcements
    
    
    def is_guild_thread_public(self):
        """
        Returns whether the channel is a guild public thread.
        
        Returns
        -------
        is_guild_thread_public : `bool`
        """
        return self.type is ChannelType.guild_thread_public
    
    
    def is_guild_thread_private(self):
        """
        Returns whether the channel is a guild private thread.
        
        Returns
        -------
        is_guild_thread_private : `bool`
        """
        return self.type is ChannelType.guild_thread_private
    
    
    def is_guild_stage(self):
        """
        Returns whether the channel is a guild stage channel.
        
        Returns
        -------
        is_guild_stage : `bool`
        """
        return self.type is ChannelType.guild_stage
    
    
    def is_guild_directory(self):
        """
        Returns whether the channel is a guild directory channel.
        
        Returns
        -------
        is_guild_directory : `bool`
        """
        return self.type is ChannelType.guild_directory
    
    
    def is_guild_forum(self):
        """
        Returns whether the channel is a guild forum channel.
        
        Returns
        -------
        is_guild_forum : `bool`
        """
        return self.type is ChannelType.guild_forum



__all__ = ()

import sys
from importlib.machinery import SourceFileLoader

from scarletio import include

from .module_proxy_type import PluginModuleProxyType
from .utils import create_module_from_spec

import_plugin = include('import_plugin')


class PluginSourceLoader(SourceFileLoader):
    """
    `SourceFileLoader` subclass used for plugin files.
    
    Attributes
    ----------
    _module : `None`, ``ModuleType``
        The cached module if any.
    _module_proxy : `None`, ``PluginModuleProxyType``
        Module proxy proxying the created module through an plugin.
    name : `str`
        The name of the represented module.
    path : `str`
        Path to the represented module.
    """
    def __init__(self, name, path):
        """
        Initialises the source file loader.
        
        Parameters
        ----------
        name : `str`
            The name of the represented module.
        path : `str`
            Path to the represented module.
        """
        SourceFileLoader.__init__(self, name, path)
        self._module = None
        self._module_proxy = None
    
    
    def create_module(self, spec):
        """
        Creates the module of the plugin source loader.
        
        Returns
        -------
        module : ``PluginModuleProxyType``
        """
        module_proxy = self._module_proxy
        if module_proxy is None:
            module_proxy = PluginModuleProxyType(spec)
            self._module_proxy = module_proxy
        
        module = self._module
        if module is None:
            module = create_module_from_spec(spec)
            self._module = module
        
        sys.modules[self.name] = module_proxy
        
        return module_proxy
    
    
    def exec_module(self, module):
        """
        Executes the given module in the context.
        
        Parameters
        ----------
        module : ``ModuleType``, ``PluginModuleProxyType``
            The module to execute.
            
            > If called from importlib, `module` is passed as ``PluginModuleProxyType``.
        """
        if isinstance(module, PluginModuleProxyType):
            module_proxy = self._module_proxy
            if (module_proxy is None):
                raise RuntimeError(
                    f'Module `{self.name}` is not created yet, but some why called with {module!r}.'
                )
            
            if module is not module_proxy:
                raise RuntimeError(
                    f'`module` must be `{module_proxy!r}, got {module!r}.'
                )
            
            import_plugin(self.name)
        else:
            SourceFileLoader.exec_module(self, self._module)



__all__ = ('Team', )

from ...bases import DiscordEntity, ICON_TYPE_NONE, IconSlot
from ...core import TEAMS
from ...http import urls as module_urls
from ...user import ClientUserBase, ZEROUSER, create_partial_user_from_id

from ..team_member import TeamMember, TeamMembershipState

from .fields import (
    parse_id, parse_members, parse_name, parse_owner_id, put_id_into, put_members_into, put_name_into,
    put_owner_id_into, validate_id, validate_members, validate_name, validate_owner_id
)


PRECREATE_FIELDS = {
    'members': ('members', validate_members),
    'name': ('name', validate_name),
    'owner_id': ('owner_id', validate_owner_id),
}


class Team(DiscordEntity, immortal = True):
    """
    Represents a Team on the Discord developer portal.
    
    Attributes
    ----------
    icon_hash : `int`
        The team's icon's hash as `uint128`. Defaults to `0`.
    icon_type : `IconType`
        The team's icon's type.
    id : `int`
        The unique identifier number of the team.
    members : `list` of `TeamMember`
        The members of the team. Includes invited members as well.
    name : `str`
        The teams name.
    owner_id : `int`
        The team's owner's id.
    
    Notes
    -----
    Team objects support weakreferencing.
    """
    __slots__ = ('members', 'name', 'owner_id',)
    
    icon = IconSlot(
        'icon',
        'icon',
        module_urls.team_icon_url,
        module_urls.team_icon_url_as,
        add_updater = False,
    )
    
    
    def __new__(cls, *, icon = ..., members = ..., name = ..., owner_id = ...):
        """
        Creates a partial team instance.
        
        Parameters
        ----------
        icon : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The team's icon.
        
        members : `None`, `iterable` of ``TeamMember``, Optional (Keyword only)
            The members of the team.
        
        name : `str`, Optional (Keyword only)
            The team's name.
        
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The team's owner's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # icon
        if icon is ...:
            icon = None
        else:
            icon = cls.icon.validate_icon(icon, allow_data = True)
        
        # members
        if members is ...:
            members = None
        else:
            members = validate_members(members)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # owner_id
        if owner_id is ...:
            owner_id = 0
        else:
            owner_id = validate_owner_id(owner_id)
        
        self = object.__new__(cls)
        self.id = 0
        self.icon = icon
        self.members = members
        self.name = name
        self.owner_id = owner_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new team from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Team data.
        
        Returns
        -------
        team : ``Team``
        """
        team_id = parse_id(data)
        try:
            self = TEAMS[team_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = team_id
        
        self._set_attributes(data)
        return self
    
    
    @classmethod
    def _create_empty(cls, team_id):
        """
        Creates a new team instance with it's attribute set to their default values.
        
        Parameters
        ----------
        team_id : `int`
            The team's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.id = team_id
        self.icon_hash = 0
        self.icon_type = ICON_TYPE_NONE
        self.members = None
        self.name = ''
        self.owner_id = 0
        return self
    
    
    @classmethod
    def precreate(cls, team_id, **keyword_parameters):
        """
        Creates a cached team instance.
        
        Parameters
        ----------
        team_id : `int`
            The team's identifier.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters.
        
        Other Parameters
        ----------------
        icon : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The team's icon.
        
        members : `None`, `iterable` of ``TeamMember``, Optional (Keyword only)
            The members of the team.
        
        name : `str`, Optional (Keyword only)
            The team's name.
        
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The team's owner's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        team_id = validate_id(team_id)

        if keyword_parameters:
            processable = []
            
            # icon
            icon = cls.icon.parse_from_keyword_parameters(keyword_parameters)
            if (icon is not None):
                processable.append(('icon', icon))
            
            extra = None
            
            while keyword_parameters:
                field_name, field_value = keyword_parameters.popitem() 
                try:
                    attribute_name, validator = PRECREATE_FIELDS[field_name]
                except KeyError:
                    if extra is None:
                        extra = {}
                    extra[field_name] = field_value
                    continue
                
                attribute_value = validator(field_value)
                processable.append((attribute_name, attribute_value))
                continue
                
            if (extra is not None):
                raise TypeError(
                    f'Unused or unsettable keyword parameters: {extra!r}.'
                )
        
        else:
            processable = None
        
        try:
            self = TEAMS[team_id]
        except KeyError:
            self = cls._create_empty(team_id)
            TEAMS[team_id] = self
        else:
            if (not self.partial):
                return self
        
        if (processable is not None):
            for item in processable:
                setattr(self, *item)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the team into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        if include_internals:
            put_id_into(self.id, data, defaults)
        
        type(self).icon.put_into(self.icon, data, defaults, as_data = not include_internals)
        put_name_into(self.name, data, defaults)
        put_members_into(self.members, data, defaults, include_internals = include_internals)
        put_owner_id_into(self.owner_id, data, defaults)
        
        return data
    
    
    def to_data_user(self):
        """
        Converts the team into a json serializable object as it would be a user.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        team_id = self.id
        
        # There are more fields, but who cares.
        return {
            'id': str(team_id),
            'username': f'team{team_id}',
            'avatar': None,
            'discriminator': '0000',
            'flags': 1024,
        }
    
    
    def _set_attributes(self, data):
        """
        Sets the team's attributes from the given data (except ``.id``).
        
        Parameters
        ----------
        data : `None`, `dict` of (`str`, `object`) items
            Team dat.
        """
        self._set_icon(data)
        self.name = parse_name(data)
        self.members = parse_members(data)
        self.owner_id = parse_owner_id(data)
    
    
    def __repr__(self):
        """Returns the team's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' id = ')
        repr_parts.append(repr(self.id))
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the team's hash value."""
        team_id = self.id
        if team_id:
            return team_id
        
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Calculates the channel's hash based on their fields.
        
        This method is called by ``.__hash__`` if the channel has no ``.channel_id`` set.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # icon
        hash_value ^= hash(self.icon)
        
        # members
        members = self.members
        if (members is not None):
            hash_value ^= len(members) << 8
            
            for member in members:
                hash_value ^= hash(member)
        
        # name
        hash_value ^= hash(self.name)
        
        # owner_id
        hash_value ^= hash(self.owner_id)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two teams are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two teams are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether self is equal to other. Other must be same type as self.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance.
        
        Returns
        -------
        is_equal : `bool`
        """
        self_id = self.id
        other_id = other.id
        if self_id and other_id:
            return (self_id == other_id)
        
        # icon
        if self.icon != other.icon:
            return False
        
        # members
        if self.members != other.members:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # owner_id
        if self.owner_id != other.owner_id:
            return False
        
        return True
    
    
    @property
    def owner(self):
        """
        Returns the team's owner.
        
        Returns
        -------
        owner : ``ClientUserBase``
        """
        owner_id = self.owner_id
        if owner_id:
            owner = create_partial_user_from_id(owner_id)
        else:
            owner = ZEROUSER
        
        return owner
    
    
    def iter_members(self):
        """
        Iterates over the team members of team.
        
        This method is an iterable generator.
        
        Yields
        ------
        team_member : ``TeamMember``
        """
        members = self.members
        if (members is not None):
            yield from members
    
    
    @property
    def invited(self):
        """
        A list of the invited users to the team.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        target_state = TeamMembershipState.invited
        return [team_member.user for team_member in self.iter_members() if team_member.state is target_state]
    
    
    @property
    def accepted(self):
        """
        A list of the users, who accepted their invite to the team.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        target_state = TeamMembershipState.accepted
        return [team_member.user for team_member in self.iter_members() if team_member.state is target_state]
    
    
    def copy(self):
        """
        Copies the team.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.id = 0
        new.icon_hash = self.icon_hash
        new.icon_type = self.icon_type
        members = self.members
        if (members is not None):
            members = (*members, )
        new.members = members
        new.name = self.name
        new.owner_id = self.owner_id
        return new
    
    
    def copy_with(self, *, icon = ..., members = ..., name = ..., owner_id = ...):
        """
        Copies the team with the defined fields.
        
        Parameters
        ----------
        icon : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The team's icon.
        
        members : `None`, `iterable` of ``TeamMember``, Optional (Keyword only)
            The members of the team.
        
        name : `str`, Optional (Keyword only)
            The team's name.
        
        owner_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The team's owner's identifier.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # icon
        if icon is ...:
            icon = self.icon
        else:
            icon = type(self).icon.validate_icon(icon, allow_data = True)
        
        # members
        if members is ...:
            members = self.members
            if (members is not None):
                members = (*members,)
        else:
            members = validate_members(members)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # owner_id
        if owner_id is ...:
            owner_id = self.owner_id
        else:
            owner_id = validate_owner_id(owner_id)
        
        new = object.__new__(type(self))
        new.id = 0
        new.icon = icon
        new.members = members
        new.name = name
        new.owner_id = owner_id
        return new
    
    
    @property
    def partial(self):
        """
        Returns whether the team is partial.
        
        Returns
        -------
        partial : `bool
        """
        return (self.id == 0)



__all__ = ('CommandsV2SnapshotType',)

from scarletio import copy_docs

from ..plugin_loader import BaseSnapshotType
from ..plugin_loader.snapshot.helpers import  _get_set_difference, _merge_set_groups

from .command_processor import CommandProcessor


class CommandsV2SnapshotType(BaseSnapshotType):
    """
    Commands extension snapshot.
    
    Attributes
    ----------
    _client_reference : `WeakReferer` to `Client`
        Weakreference to the owner client instance.
    _is_difference : `bool`
        Whether the snapshot is a difference.
    added_categories : `None`, `set` of ``Category``
        Added categories.
    added_commands : `None`, `set` of ``Command``
        Added commands.
    removed_categories : `None`, `set` of ``Category``
        Removed categories.
    removed_commands : `None`, `set` of ``Command``
        Removed commands.
    """
    __slots__ = ('added_categories', 'added_commands', 'removed_categories', 'removed_commands',)
    
    @copy_docs(BaseSnapshotType.__new__)
    def __new__(cls, client):
        command_processor = getattr(client, 'command_processor', None)
        if (command_processor is None) or (not isinstance(command_processor, CommandProcessor)):
            added_categories = None
            added_commands = None
        
        else:
            category_name_to_category = command_processor.category_name_to_category
            if category_name_to_category:
                added_categories = set(category_name_to_category.values())
            else:
                added_categories = None
            
            commands = command_processor.commands
            if commands:
                added_commands = commands.copy()
            else:
                added_commands = None
        
        self = BaseSnapshotType.__new__(cls, client)
        
        self.added_categories = added_categories
        self.added_commands = added_commands
        self.removed_categories = None
        self.removed_commands = None
        
        return self
    

    @copy_docs(BaseSnapshotType.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        client = self.client
        if (client is not None):
            repr_parts.append(' of ')
            repr_parts.append(repr(client))
        
        added_categories = self.added_categories
        if (added_categories is not None):
            repr_parts.append(', added_categories=')
            repr_parts.append(repr(added_categories))
        
        removed_categories = self.removed_categories
        if (removed_categories is not None):
            repr_parts.append(', removed_categories=')
            repr_parts.append(repr(removed_categories))
        
        added_commands = self.added_commands
        if (added_commands is not None):
            repr_parts.append(', added_commands=')
            repr_parts.append(repr(added_commands))
        
        removed_commands = self.removed_commands
        if (removed_commands is not None):
            repr_parts.append(', removed_commands=')
            repr_parts.append(repr(removed_commands))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    @copy_docs(BaseSnapshotType._extract)
    def _extract(self, other):
        added_categories, removed_categories = _get_set_difference(
            _merge_set_groups(self.added_categories, other.removed_categories),
            _merge_set_groups(other.added_categories, self.removed_categories),
        )
        
        added_commands, removed_commands = _get_set_difference(
            _merge_set_groups(self.added_commands, other.removed_commands),
            _merge_set_groups(other.added_commands, self.removed_commands),
        )
        
        new = BaseSnapshotType._extract(self, other)
        
        new.added_categories = added_categories
        new.added_commands = added_commands
        new.removed_categories = removed_categories
        new.removed_commands = removed_commands
        
        return new
    
    
    @copy_docs(BaseSnapshotType.revert)
    def revert(self):
        BaseSnapshotType.revert(self)
        
        client = self.client
        if client is None:
            return
        
        command_processor = getattr(client, 'command_processor', None)
        if (command_processor is None) or (not isinstance(command_processor, CommandProcessor)):
            return
        
        added_commands = self.added_commands
        if (added_commands is not None):
            for command in added_commands:
                command_processor._remove_command(command)
                
        removed_commands = self.removed_commands
        if (removed_commands is not None):
            for command in removed_commands:
                command_processor._add_command(command)
        
        
        added_categories = self.added_categories
        if (added_categories is not None):
            for category in added_categories:
                command_processor._remove_category(category)
        
        
        removed_categories = self.removed_categories
        if (removed_categories is not None):
            for category in removed_categories:
                command_processor._add_category(category)
    
    
    @copy_docs(BaseSnapshotType.__bool__)
    def __bool__(self):
        if (self.added_categories is not None):
            return True
        
        if (self.added_commands is not None):
            return True
        
        if (self.removed_categories is not None):
            return True
        
        if (self.removed_commands is not None):
            return True
        
        return False



__all__ = ()

from scarletio import include

from ....bases import maybe_snowflake

from ..preinstanced import PermissionOverwriteTargetType


Role = include('Role')
ClientUserBase = include('ClientUserBase')


def validate_target(target):
    """
    Validates the `target` field of ``PermissionOverwrite``. target is a mixed representation of `target_id` and
    `target_type`.
    
    Parameters
    ----------
    target : ``Role``, ``ClientUserBase``, `int`
        The permission overwrite's target or it's identifier.
    
    Returns
    -------
    target_id : `int`
        The permission overwrite target's identifier.
    target_type : ``PermissionOverwriteTargetType``
        The permission overwrite's type.
    
    Raises
    ------
    TypeError
        - If `target`'s type is unexpected.
    """
    # target_id
    if isinstance(target, Role):
        target_id = target.id
        target_type = PermissionOverwriteTargetType.role
    
    elif isinstance(target, ClientUserBase):
        target_id = target.id
        target_type = PermissionOverwriteTargetType.user
    
    else:
        target_id = maybe_snowflake(target)
        if (target_id is None):
            raise TypeError(
                f'`target` can be `int`, `{Role.__name__}`, `{ClientUserBase.__name__}`, got '
                f'{target.__class__.__name__}; {target!r}.'
            )
        
        target_type = PermissionOverwriteTargetType.unknown
    
    return target_id, target_type


def put_target_into(target, data, defaults, *, include_internals = False):
    """
    Puts the `target` field of ``PermissionOverwrite`` to the given data.
    
    The `target` value must come from ``validate_target``.
    
    Parameters
    ----------
    target : `tuple` (`int`, ``PermissionOverwriteTargetType``)
        Permission overwrite `target-id`, `target_type` pair.
    data : `dict` of (`str`, `Any`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    include_internals : `bool` = `False`, Optional (Keyword only)
        Whether internal fields should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `Any`) items
    """
    target_id, target_type = target
    
    if include_internals:
        data['id'] = str(target_id)
    
    if (target_type is not PermissionOverwriteTargetType.unknown):
        data['type'] = target_type.value
    
    return data



__all__ = ('ApplicationCommandOptionChoice',)

from enum import Enum

from scarletio import RichAttributeErrorBaseType

from ..localization.helpers import get_localized_length, localized_dictionary_builder
from ..localization.utils import build_locale_dictionary, destroy_locale_dictionary, hash_locale_dictionary

from .constants import (
    APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MAX, APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MIN,
    APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MAX, APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MIN
)
from .helpers import apply_translation_into


def _assert__application_command_option_choice__name__length(name):
    """
    Asserts the `name` parameter's length of ``ApplicationCommandOptionChoice.__new__`` method.
    
    Parameters
    ----------
    name : `str`
        The application command option choice's name.
    
    Raises
    ------
    AssertionError
        - if `name`'s length is out of the expected range.
    """
    name_length = len(name)
    if (
        name_length < APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MIN or
        name_length > APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MAX
    ):
        raise AssertionError(
            f'`name` length can be in range '
            f'[{APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MIN}:{APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MAX}], '
            f'got {name_length!r}; {name!r}.'
        )
    
    return True


def _assert__application_command_option_choice__value__length(value):
    """
    Asserts the `value` parameter's length of ``ApplicationCommandOptionChoice.__new__`` method.
    
    Parameters
    ----------
    value : `int`, `float`, `str`
        The application command option choice's value.
    
    Raises
    ------
    AssertionError
        - if `value`'s length is out of the expected range.
    """
    if isinstance(value, str):
        value_length = len(value)
        if (
            value_length < APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MIN or
            value_length > APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MAX
        ):
            raise AssertionError(
                f'`value` length` can be in range '
                f'[{APPLICATION_COMMAND_CHOICE_VALUE_LENGTH_MIN}:{APPLICATION_COMMAND_CHOICE_NAME_LENGTH_MAX}]'
                f'got {value_length!r}; {value!r}.'
            )
    
    return True


class ApplicationCommandOptionChoice(RichAttributeErrorBaseType):
    """
    A choice of a ``ApplicationCommandOption``.
    
    Attributes
    ----------
    name : `str`
        The choice's name. It's length can be in range [1:100].
    name_localizations : `None`, `dict` of (``Locale``, `str`) items
        Localized names of the choice.
    value : `str`, `int`, `float`
        The choice's value.
    """
    __slots__ = ('name', 'name_localizations', 'value')
    
    def __new__(cls, name, value=None, *, name_localizations=None):
        """
        Creates a new ``ApplicationCommandOptionChoice`` with the given parameters.
        
        Parameters
        ----------
        name : `str`, `Enum`
            The choice's name. It's length can be in range [1:100].
        
        value : `None`, `str`, `int`, `float`, `Enum` = `None`, Optional
            The choice's value.
            
            Defaults to `name` parameter if not given.
        
        name_localizations : `None`, `dict` of ((`str`, ``Locale``), `str`) items,
                (`list`, `set`, `tuple`) of `tuple` ((`str`, ``Locale``), `str`) = `None`, Optional (Keyword only)
            Localized names of the choice.
        
        Raises
        ------
        TypeError
            - If `name_localizations`'s or any of it's item's type is incorrect.
            - If `name`'s type is incorrect.
            - If `value`'s type is incorrect.
        ValueError
            - If `name_localizations` has an item with incorrect structure.
        AssertionError
            - If `name`'s length is out of range [1:100].
            - If `value` is `str` and it's length is out of range [0:100].
        """
        if isinstance(name, str):
            choice_name = name
        
        elif isinstance(name, Enum):
            if value is None:
                choice_name = name.name
                
                if not isinstance(choice_name, str):
                    raise TypeError(
                        f'`{Enum.__name__}` choice\'s `.name` should be `str`, '
                        f'got {choice_name.__class__.__name__}; {choice_name!r}; name={name!r}; value={value!r}..'
                    )
                    
                value = name.value
            else:
                choice_name = name.value
                
                if not isinstance(choice_name, str):
                    raise TypeError(
                        f'`{Enum.__name__}` choice\'s `.value` should be `str`, '
                        f'got {choice_name.__class__.__name__}; {choice_name!r}; name={name!r}; value={value!r}.'
                    )
        
        else:
            raise TypeError(
                f'`name` can be `str`, `{Enum.__name__}`, got {name.__class__.__name__}; {name!r}.'
            )
        
        assert _assert__application_command_option_choice__name__length(choice_name)
        
        # name_localizations
        name_localizations = localized_dictionary_builder(name_localizations, 'name_localizations')
        
        # value
        if value is None:
            choice_value = choice_name
        
        elif isinstance(value, (str, int, float)):
            choice_value = value
        
        elif isinstance(value, Enum):
            choice_value = value
            
            if not isinstance(choice_value, (str, int, float)):
                raise TypeError(
                    f'`{Enum.__name__}` choice\'s `.value` can be `str`, `int`, `float`, '
                    f'got {choice_value.__class__.__name__}; {choice_value!r}; value={value!r}.'
                )
            
        else:
            raise TypeError(
                f'`value` type can be `str`, `int`, `float`, `{Enum.__name__}`, '
                f'got {value.__class__.__name__}; {value!r}.'
            )
        
        assert _assert__application_command_option_choice__value__length(choice_value)
        
        self = object.__new__(cls)
        self.name = choice_name
        self.name_localizations = name_localizations
        self.value = choice_value
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``ApplicationCommandOptionChoice`` from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            The received application command option choice data.
        
        Returns
        -------
        self : ``ApplicationCommandOptionChoice``
            The created choice.
        """
        # name
        name = data['name']
        
        # name_localizations
        name_localizations = build_locale_dictionary(data.get('name_localizations', None))
        
        # value
        value = data['value']
        
        self = object.__new__(cls)
        self.name = name
        self.name_localizations = name_localizations
        self.value = value
        return self
    
    
    def to_data(self):
        """
        Converts the application command option choice to a json serializable object.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        
        # name
        data['name'] = self.name
        
        # name_localizations
        data['name_localizations'] = destroy_locale_dictionary(self.name_localizations)
        
        # value
        data['value'] = self.value
        
        return data
    
    
    def copy(self):
        """
        Copies the ``ApplicationCommandOptionChoice``.
        
        Returns
        -------
        new : ``ApplicationCommandOptionChoice``
        """
        new = object.__new__(type(self))
        
        # name
        new.name = self.name
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            name_localizations = name_localizations.copy()
        new.name_localizations = name_localizations
        
        # value
        new.value = self.value
        
        return new
    
    
    def __repr__(self):
        """Returns the application command option choice's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # name
        repr_parts.append(' name=')
        repr_parts.append(repr(self.name))
        
        # value
        repr_parts.append(', value=')
        repr_parts.append(repr(self.value))
        
        # Extra fields: `.name_localizations`
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            repr_parts.append(', name_localizations=')
            repr_parts.append(repr(name_localizations))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two choices are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # name
        if self.name != other.name:
            return False
        
        # name_localizations
        if self.name_localizations != other.name_localizations:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def __len__(self):
        """Returns the application command option choice's length."""
        length = 0
        
        # name & name_localizations
        length += get_localized_length(self.name, self.name_localizations)
        
        # value
        value = self.value
        if isinstance(value, str):
            length += len(value)
        
        return length
    
    def __hash__(self):
        """Returns the application command option choice's representation."""
        hash_value = 0
        
        # name
        hash_value ^= hash(self.name)
        
        # name_localizations
        name_localizations = self.name_localizations
        if (name_localizations is not None):
            hash_value ^= hash_locale_dictionary(name_localizations)
        
        # value
        # Do not hash `.value` of equals to `.name`
        value = self.value
        if (not isinstance(value, str)) or (value != self.name):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def apply_translation(self, translation_table, replace=False):
        """
        Applies translation from the given nested dictionary to the application command option choice.
        
        Parameters
        ----------
        translation_table : `None`, `dict` of ((``Locale``, `str`),
                (`None`, `dict` (`str`, (`None`, `str`)) items)) items
            Translation table to pull localization. from.
        replace : `bool` = `False`, Optional
            Whether actual translation should be replaced.
        """
        if translation_table is None:
            return
        
        # name
        self.name_localizations = apply_translation_into(
            self.name,
            self.name_localizations,
            translation_table,
            replace,
        )
